/***************************************************************************
mondoarchive.c  
-------------------
begin                : Fri Apr 19 16:40:35 EDT 2002
copyright						 : (C) 2002 Mondo  Hugo Rabson
email								 : Hugo Rabson <hugo@firstlinux.net>
edited by            : by Stan Benoit 4/2002
email                : troff@nakedsoul.org
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

/***************************************************************************
UPDATE LOG

07/13
- if not '-o' then insist on vfat-friendly kernel

07/10
- removed "the rest are" comment from log

06/29
- added signal-trapping

06/21
- added '-W' (won't make bootable CD or floppies) flag

06/05
- added missing 'j++' to -s switch's code

06/03
- expanded -s switch

05/23
- added -u switch

05/15
- if dvdrecord exists and cdrecord doesn't then use
  dvdrecord instead of cdrecord

04/01 - 04/30
- if CD-ROM is supermounted then unmount, run, and remount
- replace MONDO_VERSION #define with VERSION from ../config.h
- if CD-ROM is mounted at start then abort
- improved homedir-locating code
- improved "backup will occupy N CD's" calcuation a bit
- added -m (manual CD tray) flag

03/01 - 03/31
- read 1MB from tape instead of 1KB
- record /proc/cpuinfo and uname -r in logfile
- record hostname (Krzysztof Dubowik)
- if '-L' then test for lzop; abort if not present
- tell user to use 'su -' instead of 'su' if they are dumbasses

02/01 - 02/28
- added lots of cdstream-related code
- change tape-testing dd from 64K to 4K
- add tmp.mondo.NNNN to tempdir path, if tempdir is specified by user
- add mondo.scratch.NNN to scratchdir path, if scratchdir is specified by user
- fixed typo - should say "man mondarchive"
- if errors apparently occurred, don't just say, "Errors occurred": ask the
  user to check the logs as a precaution
- insist on cdrecord being installed if -c or -w
- replaced '&> /dev/null' with '> /dev/null 2> /dev/null'
- don't try to write floppies if images don't exist
- don't force user to specify tape size if verifying only
- NFS-related fixes
- don't log g_mondo_home
- estimate, log + print the number of CD's/tapes required if backing up
- find Mondo's homedir automatically
- don't try to use 'dd' to check the validity of device unless it is a tape
- added /tmp/mondo-restore.cfg file, with a view to replacing
  the various boot-time BLAH-BLAH and FOO-BAR files in /tmp and /

01/01 - 01/31
- minor improvements to parameter-related feedback/errors
- ported remaining shell script to C
- handles files >2GB in size

*******************************************************************/
#include "my-stuff.h"
#include "mondoarchive.h"
#include "mondostructures.h"

#include "../config.h"


/** file pointer **************************************************/
FILE *g_tape_stream = NULL;

/** long long *****************************************************/
long long g_tape_posK = 0;

/** long **********************************************************/
long g_noof_sets = 0;

/** bool******** **************************************************/
bool g_debugging = FALSE;
bool g_running_live = FALSE;
bool g_cd_recovery = FALSE;
bool g_ISO_mode = FALSE;

/*int current_backup_media;*/

/** newt ***********************************************************/
newtComponent g_progressForm;


/** dummies functions **********************************************/

/*************************************************************************
 * what_number_cd_is_this() -- Hugo Rabson 											         *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * returns:                                                              *
 *************************************************************************/
int
what_number_cd_is_this (void)
{
  log_it ("FOO1");
  exit (100);
}

/*************************************************************************
 * load_mountlist() -- Hugo Rabson            									         *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * returns:                                                              *
 *************************************************************************/
int
load_mountlist (struct mountlist_itself *a, char *b)
{
  log_it ("FOO2");
  exit (100);
}

/*************************************************************************
 * load_raidtab_into_raidlist() -- Hugo Rabson 									         *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * returns:                                                              *
 *************************************************************************/
int
load_raidtab_into_raidlist (struct raidlist_itself *a, char *b)
{
  log_it ("FOO101");
  exit (101);
}

/*************************************************************************
 * sort_mountlist_by_device() -- Hugo Rabson 	  								         *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * returns:                                                              *
 *************************************************************************/
void
sort_mountlist_by_device (struct mountlist_itself *a, bool b)
{
  log_it ("FOO4");
  exit (100);
}

/*************************************************************************
 * sort_mountlist_entries() -- Hugo Rabson 											         *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * returns:                                                              *
 *************************************************************************/
void
sort_mountlist_entries (struct mountlist_itself *a, bool b)
{
  log_it ("FOO5");
  exit (100);
}


/*************************************************************************
 * swap_mountlist_entries() -- Hugo Rabson    									         *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * returns:                                                              *
 *************************************************************************/
void
swap_mountlist_entries (struct mountlist_itself *a, int b, int c)
{
  log_it ("FOO6");
  exit (100);
}

/*************************************************************************
 * copy_mondo_and_mindi_stuff() -- Hugo Rabson 									         *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * returns:                                                              *
 *************************************************************************/
void
copy_mondo_and_mindi_stuff_to_scratchdir (struct s_bkpinfo *bkpinfo)
{
	/** Char buffers ***/
  char command[MAX_STR_LEN];
  char tmp[MAX_STR_LEN];

  mvaddstr_and_log_it (g_currentY, 0,
		       "Copying Mondo's core files to the scratch directory");

  sprintf (command, "cp --parents -pRdf %s %s", g_mondo_home,
	   bkpinfo->scratchdir);

  if (run_program_and_log_output (command))
    {
      fatal_error ("Failed to copy Mondo's stuff to scratchdir");
    }

  sprintf (command, "cp -f %s/LAST-FILELIST-NUMBER %s", bkpinfo->tmpdir,
	   bkpinfo->scratchdir);

  if (run_program_and_log_output (command))
    {
      fatal_error ("Failed to copy LAST-FILELIST-NUMBER to scratchdir");
    }

  strcpy (tmp,
	  call_program_and_get_last_line_of_output ("which mondorestore"));

  sprintf (command, "cp -f %s %s", tmp, bkpinfo->tmpdir);

  if (run_program_and_log_output (command))
    {
      fatal_error ("Failed to copy mondorestore to tmpdir");
    }

  sprintf (command, "hostname > %s/HOSTNAME", bkpinfo->scratchdir);

  system (command);

  mvaddstr_and_log_it (g_currentY++, 74, "Done.");
}

/*************************************************************************
 * eval_call() -- Hugo Rabson 				                 *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * returns:                                                              *
 *************************************************************************/
int
eval_call (struct s_bkpinfo *bkpinfo,
	   char *basic_call, char *isofile,
	   int cd_no, char *logstub, char *what_i_am_doing)
{
  	
	/** int's  ****/
  int retval = 0;
  int res = 0;
  int blanklines = 0;
  int jj = 0;
  int i = 0;

  /** File pointer ****/
  FILE *fin;

	/** buffers      ****/
  char midway_call[MAX_STR_LEN];
  char ultimate_call[MAX_STR_LEN];
  char tmp[MAX_STR_LEN];
  char command[MAX_STR_LEN];
  char incoming[MAX_STR_LEN];
  char old_stderr[MAX_STR_LEN];
  char stderr_fname[MAX_STR_LEN];
  char lockfile[MAX_STR_LEN];
  char cd_number_str[MAX_STR_LEN];
  /* char popup_str[MAX_STR_LEN];  removed 19 apr 02 stan benoit */
  char *p;

  old_stderr[0] = '\0';
/*************   End Varibles ***************************************/

  sprintf (cd_number_str, "%d", cd_no);
  sprintf (stderr_fname, "%s/stderr.txt", bkpinfo->tmpdir);
  resolve_naff_tokens (midway_call, basic_call, isofile, "_ISO_");
  resolve_naff_tokens (tmp, midway_call, cd_number_str, "_CD#_");
  resolve_naff_tokens (ultimate_call, tmp, stderr_fname, "_ERR_");
  sprintf (lockfile,
	   call_program_and_get_last_line_of_output
	   ("mktemp -q /tmp/mondo.XXXXXXXX"));
  unlink(stderr_fname);
  /*  if (!g_text_mode)
    {
  */
      sprintf (ultimate_call, "%s 2>> %s", ultimate_call, stderr_fname);
      //    } 

  sprintf (command, "echo hi > %s ; %s ; res=$?; rm -f %s; exit $res",
	   lockfile, ultimate_call, lockfile);

  open_evalcall_form (what_i_am_doing);
  strcpy (tmp, command);
  if (bkpinfo->manual_cd_tray)
    {
      p = strstr (tmp, "2>>");
      sprintf (p, "   ");
      while(*p==' ') { p++; }
      for (; *p != ' '; p++)
	{
	  *p = ' ';
	}

      strcpy(command, tmp);
      if (!g_text_mode)
	{
	  newtFinished ();
	}
      log_it (command);
      retval += system (command);
      //      clear();
      setup_newt_stuff ();
      if (retval)
	{
	  popup_and_OK ("Failed to burn CD properly.");
	}
      return (retval);
    }
  /* if text mode then do the above & RETURN; if not text mode, do this... */

  sprintf (tmp, "NB: CD-burning command (FINAL) = %s", command);
  log_it(tmp);

  fin = popen (command, "r");
  if (!fin)
    {
      sprintf (tmp, "Failed utterly to call '%s'", command);
      log_to_screen (tmp);
      return (1);
    }
  for (i = 0; i < 5 && !does_file_exist (lockfile); sleep (1), i++);
  for (i = 0; does_file_exist (lockfile); sleep (1), i = (i + 1) % 10)
    {
      /*
      if (i==0)
	{
	  log_it("Still counting...");
      */
	  log_file_end_to_screen (stderr_fname, "");
	  jj = grab_percentage_from_last_line_of_file (stderr_fname);
	  if (jj>0 && i==0)
	    {
	      sprintf(tmp,"%d%% complete. Still working...", jj);
	      log_it(tmp);
	    }
	  /*
	  sprintf(tmp,"jj = %d",jj);
	  log_it(tmp);
	  */
	  update_evalcall_form (jj);
	  //	}
    }
  log_it("Done counting.");

  for (fgets (incoming, MAX_STR_LEN, fin); !feof (fin);
       fgets (incoming, MAX_STR_LEN, fin))
    {
      incoming[78] = '\0';
      while (incoming[strlen (incoming) - 1] < 32 && strlen(incoming)>0)
	{
	  incoming[strlen (incoming) - 1] = '\0';
	}
      if (strlen (incoming) == 0)
	{
	  blanklines++;
	}
      log_to_screen (incoming);
    }

  log_it("About to close...");

  if (pclose (fin))
    {
      res = 1;
    }
  else
    {
      res = 0;
    }
  retval += res;
  close_evalcall_form ();
  log_it("Closed. Cool.");
  if (does_file_exist(stderr_fname))
    {
      sprintf (tmp, "cat %s >> %s", stderr_fname, MONDO_LOGFILE);
      system (tmp);
    }
  if (strstr (incoming, "<CR>"))
    {
      popup_and_OK(MONDO_POPMSG ); /* stan benoit 22 apr 2002 */
      retval++;
    }
  unlink (stderr_fname);
  return (retval);
}


/*************************************************************************
 * handle_incoming_parameters() -- Hugo Rabson                           *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * returns:                                                              *
 *************************************************************************/
int
handle_incoming_parameters (int argc, char *argv[], struct s_bkpinfo *bkpinfo)
{
	/** int ****/
  int res = 0;
  int retval = 0;
  int i = 0, j;

	/** buffers ****************/
  char command[MAX_STR_LEN];
  char tmp[MAX_STR_LEN];
  char flag_val[128][MAX_STR_LEN];
  bool flag_set[128];



  for (i = 0; i < 128; i++)
    {
      flag_val[i][0] = '\0';
      flag_set[i] = FALSE;
    }
  strcpy (bkpinfo->tmpdir, "/root/images/mondo");
  strcpy (bkpinfo->scratchdir, "/home");
  bkpinfo->compression_level = 3;
  bkpinfo->optimal_set_size = 5 * 1024;
  getcwd (g_startdir, MAX_STR_LEN);
  strcpy (bkpinfo->isodir, "/root/images/mondo");
  sprintf (command, "mkdir -p %s 2> /dev/null", bkpinfo->isodir);
  system (command);
  strcpy (bkpinfo->include_paths, "/");
  for(j=1; j<MAX_NOOF_MEDIA; j++) {bkpinfo->media_size[j] = 650;}	/* default */
  res = retrieve_switches_from_command_line (argc, argv, flag_val, flag_set);
  retval += res;
  if (!retval)
    {
      res = process_switches (bkpinfo, flag_val, flag_set);
      retval += res;
    }
  if (!retval)
    {
      log_it ("Switches:-");
      for (i = 0; i < 128; i++)
	{
	  if (flag_set[i])
	    {
	      sprintf (tmp, "-%c %s", i, flag_val[i]);
	      log_it (tmp);
	    }
	}
    }
  sprintf (tmp, "rm -Rf %s/tmp.mondo.*", bkpinfo->tmpdir);
  system (tmp);
  sprintf (tmp, "rm -Rf %s/mondo.scratch.*", bkpinfo->scratchdir);
  system (tmp);
  sprintf (bkpinfo->tmpdir + strlen (bkpinfo->tmpdir), "/tmp.mondo.%ld",
	   random () % 32767);
  sprintf (bkpinfo->scratchdir + strlen (bkpinfo->scratchdir),
	   "/mondo.scratch.%ld", random () % 32767);
  sprintf (tmp, "mkdir -p %s/tmpfs", bkpinfo->tmpdir);
  system (tmp);
  sprintf (tmp, "mkdir -p %s", bkpinfo->scratchdir);
  system (tmp);
  if (bkpinfo->nfs_mount[0] != '\0')
    {
      store_nfs_config (bkpinfo);
    }
  return (retval);
}







int process_the_s_switch(struct s_bkpinfo *bkpinfo, char *value)
{
      int i,j;
      char tmp[MAX_STR_LEN], *p, comment[MAX_STR_LEN];

      bkpinfo->media_size[0] = -1; /* dummy value */
      for(j=1, p=value; j<MAX_NOOF_MEDIA && strchr(p,','); j++, p=strchr(p,',')+1)
	{
	  strcpy(tmp, p);
	  *(strchr(tmp,',')) = '\0';
	  bkpinfo->media_size[j] = friendly_sizestr_to_sizelong (tmp);
          sprintf(comment, "media_size[%d] = %ld", j, bkpinfo->media_size[j]);
          log_it(comment);
	}
      for(; j<=MAX_NOOF_MEDIA; j++)
	{
	  bkpinfo->media_size[j] = friendly_sizestr_to_sizelong (p);
	}
//      bkpinfo->media_size[0] = bkpinfo->media_size[MAX_NOOF_MEDIA];
      for(j=1; j<=MAX_NOOF_MEDIA; j++)
        {
          if (bkpinfo->media_size[j]<=0) { fprintf(stderr, "You gave media #%d an invalid size\n", j); return(-1); }
        }
      return(0);
}




/*************************************************************************
 * process_switches() -- Hugo Rabson                                     *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
int
process_switches (struct s_bkpinfo *bkpinfo, char flag_val[128][MAX_STR_LEN],
		  bool flag_set[128])
{

	/** ints ****/
  int i = 0, j;
  int retval = 0;

	/** buffers ***/
  char extra_cdrom_params[MAX_STR_LEN];
  char tmp[MAX_STR_LEN], *p;
  char mondo_mkisofs_sz[MAX_STR_LEN];
  char cdr_exe[MAX_STR_LEN];

/* compulsory */
  i =
    flag_set['c'] + flag_set['i'] + flag_set['n'] + flag_set['t'] + flag_set['u'] +
    flag_set['w'] + flag_set['C'];
  if (i == 0)
    {
      retval++;
      fprintf (stderr, "You must specify the media type\n");
    }
  if (i > 1)
    {
      retval++;
      fprintf (stderr, "Please specify only one media type\n");
    }
  if (flag_set['L'] && flag_set['0'])
    {
      retval++;
      fprintf (stderr, "You cannot have 'no compression' _and_ LZOP.\n");
    }
  i = flag_set['O'] + flag_set['V'];
  if (i == 0)
    {
      retval++;
      fprintf (stderr, "Specify backup (-O), verify (-V) or both (-OV).\n");
    }
  bkpinfo->backup_data = flag_set['O'];
  bkpinfo->verify_data = flag_set['V'];
  if (flag_set['I'] && !bkpinfo->backup_data)
    {
      log_to_screen ("-I switch is ignored if just verifying");
    }
  if (flag_set['E'] && !bkpinfo->backup_data)
    {
      log_to_screen ("-E switch is ignored if just verifying");
    }
  if (flag_set['W'])
    {
      bkpinfo->nonbootable_backup = TRUE;
      log_to_screen ("Warning - you have opted for non-bootable backup");
      if (flag_set['f'] || flag_set['l'])
        {
          log_to_screen ("You don't need to specify bootloader or bootdevice");
        }
    }
  if (flag_set['I'])
    {
      strcpy (bkpinfo->include_paths, flag_val['I']);
      if (bkpinfo->include_paths[0] == '-'
	  || bkpinfo->include_paths[0] == '\0')
	{
	  retval++;
	  fprintf (stderr, "Please supply a sensible value with '-I'\n");
	}
    }
  if (flag_set['c'] || flag_set['w'] || flag_set['C'])
    {
      if (system ("which cdrecord > /dev/null 2> /dev/null") && system("which dvdrecord > /dev/null 2> /dev/null"))
	{
	  fatal_error ("Please install dvdrecord/cdrecord and try again.");
	}
      if (flag_set['C'])
	{
	  bkpinfo->using_cdstream = TRUE;
	  bkpinfo->cdrw_speed = 2;
	  if (!flag_set['L'])
	    {
	      retval++;
	      fprintf (stderr, "Please use -L with -C. It is mandatory.\n");
	    }
	}
      else
	{
	  bkpinfo->cdrw_speed =
	    atoi ((flag_set['c']) ? flag_val['c'] : flag_val['w']);
	  if (bkpinfo->cdrw_speed < 2)
	    {
	      fatal_error ("You specified a silly speed for a CD-R[W] drive");
	    }
	}
    }
  if (flag_set['t'] || flag_set['u'])
    {				/* tape size */
      bkpinfo->using_tape = TRUE;
      if (flag_set['O'])
	{
	  if (flag_set['s'])
	    {
              if (process_the_s_switch(bkpinfo, flag_val['s'])) {fatal_error("Bad -s switch");}
            }
	  else if (flag_set['u'])
    	{
	      bkpinfo->media_size[1] = 999999999; /* nearly infinite :) */
	    }
	  else
	    {
	      retval++;
	      fprintf (stderr, "Tape size not specified.\n");
	    }
	}
    }
  else
    {				/* CD size */
      bkpinfo->using_tape = FALSE;
      if (flag_set['s'])
	{
          if (process_the_s_switch(bkpinfo, flag_val['s'])) {fatal_error("Bad -s switch");}
	}
      if (flag_set['w'])
	{
	  bkpinfo->wipe_media_first = TRUE;
	}			/* CD-RW */
    }
  if (flag_set['n'])
    {
      strcpy (bkpinfo->nfs_mount, flag_val['n']);
      if (!flag_set['d'])
	{
	  strcpy (bkpinfo->nfs_remote_dir, "/");
	}
      sprintf (tmp, "mount | grep -x \"%s .*\" | cut -d' ' -f3",
	       bkpinfo->nfs_mount);
      strcpy (bkpinfo->isodir,
	      call_program_and_get_last_line_of_output (tmp));
      if (strlen (bkpinfo->isodir) < 3)
	{
	  retval++;
	  fprintf (stderr, "NFS share is not mounted. Please mount it.\n");
	}
      sprintf (tmp, "mount = %s", bkpinfo->nfs_mount);
      //      log_it(tmp);
      sprintf (tmp, "isodir= %s", bkpinfo->isodir);
      //      log_it(tmp);
      sprintf (tmp, "ls %s/%s", bkpinfo->isodir, bkpinfo->nfs_remote_dir);
      if (run_program_and_log_output (tmp))
	{
	  retval++;
	  fprintf (stderr,
		   "Are you sure directory '%s' exists in remote dir '%s'?\n",
		   bkpinfo->nfs_remote_dir, bkpinfo->nfs_mount);
	}
    }

/* optional, popular */
  if (flag_set['E'])
    {
      strcpy (bkpinfo->exclude_paths, flag_val['E']);
    }
  if (flag_set['g'])
    {
      g_text_mode = FALSE;
    }
  if (flag_set['D'])
    {
      bkpinfo->differential = TRUE;
    }
  if (flag_set['x'])
    {
      strcpy (bkpinfo->image_devs, flag_val['x']);
    }
  if (flag_set['m'])
    {
      bkpinfo->manual_cd_tray = TRUE;
    }
  if (flag_set['k'])
    {
      strcpy (bkpinfo->kernel_path, flag_val['k']);
      if (!strcmp (bkpinfo->kernel_path, "failsafe"))
	{
	  strcpy (bkpinfo->kernel_path, "FAILSAFE");
	}
      if (strcmp (bkpinfo->kernel_path, "FAILSAFE")
	  && !does_file_exist (bkpinfo->kernel_path))
	{
	  retval++;
	  fprintf (stderr,
		   "You specified kernel '%s', which does not exist\n",
		   bkpinfo->kernel_path);
	}
    }
  if (flag_set['d'])
    {				/* backup directory (if ISO/NFS) */
      if (flag_set['i'])
	{
	  strcpy (bkpinfo->isodir, flag_val['d']);
	}
      else if (flag_set['n'])
	{
	  strcpy (bkpinfo->nfs_remote_dir, flag_val['d']);
	}
      else			/* backup device (if tape/CD-R/CD-RW) */
	{
	  strcpy (bkpinfo->media_device, flag_val['d']);

	}
    }
  if (!flag_set['d'] && (flag_set['c'] || flag_set['w'] || flag_set['C']))
    {
      if (find_cdrw_device (bkpinfo->media_device))
	{
	  retval++;
	  fprintf (stderr,
		   "Tried and failed to find CD-R[W] drive automatically.\n");
	}
      else
	{
	  flag_set['d'] = TRUE;
	  strcpy (flag_val['d'], bkpinfo->media_device);
	}
    }
  if (!flag_set['d'] && !flag_set['n'] && !flag_set['C'])
    {
      retval++;
      fprintf (stderr, "Please specify the backup device/directory.\n");
    }

/* optional, obscure */
  for (i = '0'; i <= '9'; i++)
    {
      if (flag_set[i])
	{
	  bkpinfo->compression_level = i - '0';
	}			/* not '\0' but '0' */
    }
  if (flag_set['S'])
    {
      sprintf (bkpinfo->scratchdir, "%s/mondo.scratch.%ld", flag_val['S'],
	       random () % 32768);
    }
  if (flag_set['T'])
    {
      sprintf (bkpinfo->tmpdir, "%s/tmp.mondo.%ld", flag_val['T'],
	       random () % 32768);
    }
  if (flag_set['A'])
    {
      strcpy (bkpinfo->call_after_iso, flag_val['A']);
    }
  if (flag_set['B'])
    {
      strcpy (bkpinfo->call_before_iso, flag_val['B']);
    }
  if (flag_set['F'])
    {
      g_skip_floppies = TRUE;
    }
  if (flag_set['H'])
    {
      g_cd_recovery = TRUE;
    }
  if (flag_set['l'])
    {
      bkpinfo->boot_loader = flag_val['l'][0];
    }
  if (flag_set['f'])
    {
      strcpy (bkpinfo->boot_device, flag_val['f']);
    }
  if (flag_set['P'])
    {
      strcpy (bkpinfo->postnuke_tarball, flag_val['P']);
    }
  if (flag_set['L'])
    {
      bkpinfo->use_lzo = TRUE;
      if (run_program_and_log_output ("which lzop"))
	{
	  retval++;
	  fprintf (stderr,
		   "Please install LZOP. You can't use '-L' until you do.\n");
	}
    }
  if (flag_set['o'])
    {
      bkpinfo->make_cd_use_lilo = TRUE;
    }
  else
    {
      run_program_and_log_output("insmod dos");
      run_program_and_log_output("insmod fat");
      run_program_and_log_output("insmod vfat");
      if (run_program_and_log_output("grep vfat /proc/filesystems"))
        { fatal_error("Either call mondoarchive with '-o', or recompile your kernel to include VFAT support."); }
    }
/* and finally... */
  g_current_media_number = 1;
  if (bkpinfo->use_lzo)
    {
      strcpy (bkpinfo->zip_exe, "lzop");
      strcpy (bkpinfo->zip_suffix, "lzo");
    }
  else if (bkpinfo->compression_level!=0)
    {
      strcpy (bkpinfo->zip_exe, "bzip2");
      strcpy (bkpinfo->zip_suffix, "bz2");
    }
  else
    {
      bkpinfo->zip_exe[0] = bkpinfo->zip_suffix[0] = '\0';
    }
  if (flag_set['c'] || flag_set['w'])
    {
      extra_cdrom_params[0] = '\0';
      if (!bkpinfo->manual_cd_tray)
	{
	  strcat (extra_cdrom_params, "-waiti ");
	}
      if (flag_set['w'])
	{
	  strcat (extra_cdrom_params, "blank=fast ");
	}
      if (!system("which dvdrecord > /dev/null 2> /dev/null"))
	{ strcpy(cdr_exe, "dvdrecord"); }
      else
	{ strcpy(cdr_exe, "cdrecord"); }
      if (flag_set['W'])
        { strcpy(mondo_mkisofs_sz, MONDO_MKISOFS_NONBOOT); }
      else if (flag_set['o'])
        { strcpy(mondo_mkisofs_sz, MONDO_MKISOFS_REGULAR_LILO); }
      else
        { strcpy(mondo_mkisofs_sz, MONDO_MKISOFS_REGULAR_SYSLINUX); }
      if (bkpinfo->manual_cd_tray)
	{
	  sprintf (bkpinfo->call_before_iso,
		   "%s -o %s/temporary.iso . 2>> _ERR_", mondo_mkisofs_sz, bkpinfo->tmpdir);
	  sprintf (bkpinfo->call_make_iso,
		   "%s -eject -v %s fs=4m dev=%s speed=%d %s/temporary.iso",
		   cdr_exe,
		   extra_cdrom_params, bkpinfo->media_device,
		   bkpinfo->cdrw_speed, bkpinfo->tmpdir);
	}
      else
	{
	  sprintf (bkpinfo->call_make_iso,
		   "%s . 2>> _ERR_ | %s -eject %s fs=4m dev=%s speed=%d -",
		   mondo_mkisofs_sz, cdr_exe, extra_cdrom_params, bkpinfo->media_device,
		   bkpinfo->cdrw_speed);
	}
      sprintf (tmp, "NB: call_before_iso = %s\nNB: call_make_iso = %s\n", bkpinfo->call_before_iso, bkpinfo->call_make_iso);
      log_it (tmp);
    }
  if (flag_set['t'] && flag_set['O'])
    {
      sprintf (tmp,
	       "dd if=/dev/zero of=%s bs=4k count=1 > /dev/null 2> /dev/null",
	       bkpinfo->media_device);
      if (system (tmp))
	{
	  retval++;
	  fprintf (stderr,
		   "Cannot write to tape device. Is the tape set read-only?\n");
	}
    }
  //  log_it("Finished processing incoming params");
  if (retval)
    {
      fprintf (stderr, "Type 'man mondoarchive' for help.\n");
    }
  sprintf (tmp, "%s", MONDO_TMPISOS);	/* added #define 22 apr 2002 */
  if (does_file_exist (tmp))
    {
      unlink (tmp);
    }
  return (retval);
}


/*************************************************************************
 * resolve_naff_tokens() -- Hugo Rabson                                   *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * returns:                                                              *
 *************************************************************************/
void
resolve_naff_tokens (char *output, char *ip, char *value, char *token)
{
	/** Buffers ****/
  char input[MAX_STR_LEN];

	/** Pointers **/
  char *p;

  strcpy (output, ip);		/* just in case the token doesn't appear in string at all */
  for (strcpy (input, ip); strstr (input, token); strcpy (input, output))
    {
      strcpy (output, input);
      p = strstr (output, token);
      *p = '\0';
      strcat (output, value);
      p = strstr (input, token) + strlen (token);
      strcat (output, p);
    }
}



/*************************************************************************
 * retrieve_switches() -- Hugo Rabson                                    *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * returns:                                                              *
 *************************************************************************/
int
retrieve_switches_from_command_line (int argc, char *argv[],
				     char flag_val[128][MAX_STR_LEN],
				     bool flag_set[128])
{
	/** ints ***/
  int opt = 0;
  int i = 0;

	/** bools ****/
  bool bad_switches = FALSE;

  for (i = 0; i < 128; i++)
    {
      flag_val[i][0] = '\0';
      flag_set[i] = FALSE;
    }
  while ((opt =
	  getopt (argc, argv,
		  "0123456789A:B:CDE:FHI:LOP:S:T:VWc:d:f:gik:l:mn:os:tuw:x:")) !=
	 -1)
    {
      if (opt == '?')
	{
	  bad_switches = TRUE;
	  /*fprintf(stderr,"Invalid option: %c\n",optopt); */
	}
      else
	{
	  if (flag_set[optopt])
	    {
	      bad_switches = TRUE;
	      fprintf (stderr, "Switch -%c previously defined as %s\n", opt,
		       flag_val[i]);
	    }
	  else
	    {
	      flag_set[opt] = TRUE;
	      if (optarg)
		{
		  strcpy (flag_val[opt], optarg);
		}
	    }
	}
    }
  for (i = optind; i < argc; i++)
    {
      bad_switches = TRUE;
      fprintf (stderr, "Invalid arg -- %s\n", argv[i]);
    }
  return (bad_switches);
}



/*************************************************************************
 * make_list_of_files_to_ignore() -- Hugo Rabson 	                       *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * returns:                                                              *
 *************************************************************************/
void
make_list_of_files_to_ignore (char *ignorefiles_fname, char *filelist_fname,
			      char *cklist_fname)
{
	/** buffers ****/
  char command[MAX_STR_LEN];


  /* make a list of files which have changed on the disk since backup */
  log_to_screen
    ("Making a list of files which have changed on hard disk since backup\n");
  sprintf (command,
	   "%s/mondo-checksum %s %s / --verify | grep -v \"not found on filesystem\" | cut -f1 > %s",
	   g_mondo_home, filelist_fname, cklist_fname, ignorefiles_fname);
  log_it (command);
  if (system (command))
    {
      log_it ("Failed to make list of files to ignore");
    }
  exclude_nonexistent_files (ignorefiles_fname);
}


/*************************************************************************
 * store_nfs_config() -- Hugo Rabson	                                   *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * returns:                                                              *
 *************************************************************************/
void
store_nfs_config (struct s_bkpinfo *bkpinfo)
{

	/** buffers *********/
  char outfile[MAX_STR_LEN];
  char nfs_dev[MAX_STR_LEN];
  char nfs_mount[MAX_STR_LEN];
  char nfs_client_ipaddr[MAX_STR_LEN];
  char nfs_server_ipaddr[MAX_STR_LEN];
  char tmp[MAX_STR_LEN];
  char command[MAX_STR_LEN];

	/** pointers ******/
  char *p;
  FILE *fout;



  log_it ("Storing NFS configuration");
  strcpy (tmp, bkpinfo->nfs_mount);
  p = strchr (tmp, ':');
  if (!p)
    {
      fatal_error
	("NFS mount doesn't have a colon in it, e.g. 192.168.1.4:/home/nfs");
    }
  *(p++) = '\0';
  strcpy (nfs_server_ipaddr, tmp);
  strcpy (nfs_mount, p);
  sprintf (command,
	   "ifconfig | tr '\n' '#' | sed s/##// | tr '#' ' ' | tr '' '\n' | head -n1 | cut -d' ' -f1");
  strcpy (nfs_dev, call_program_and_get_last_line_of_output (command));
  sprintf (command,
	   "ifconfig | tr '\n' '#' | sed s/##// | tr '#' ' ' | tr '' '\\n' | head -n1 | tr -s '\t' ' ' | cut -d' ' -f7 | cut -d':' -f2");
  strcpy (nfs_client_ipaddr,
	  call_program_and_get_last_line_of_output (command));
  sprintf (tmp, "nfs_client_ipaddr=%s; nfs_server_ipaddr=%s; nfs_mount=%s",
	   nfs_client_ipaddr, nfs_server_ipaddr, nfs_mount);
  //  log_it(tmp);
  if (strlen (nfs_dev) < 2)
    {
      fatal_error
	("Unable to find ethN (eth0, eth1, ...) adapter via NFS mount you specified.");
    }
  sprintf (outfile, "%s/start-nfs", bkpinfo->tmpdir);
  sprintf (tmp, "outfile = %s", outfile);
  log_it (tmp);
  if (!(fout = fopen (outfile, "w")))
    {
      fatal_error ("Cannot store NFS config");
    }
  fprintf (fout, "ifconfig %s %s; # config client\n", nfs_dev,
	   nfs_client_ipaddr);
  fprintf (fout, "# ping -c1 %s; # ping server\n", nfs_server_ipaddr);
  fprintf (fout, "# mount %s -t nfs /mnt/isodir\n", bkpinfo->nfs_mount);
  fprintf (fout, "exit 0\n");
  fclose (fout);
  chmod (outfile, 777);
  system ("mkdir -p /var/cache/mondo-archive 2> /dev/null");
  sprintf (tmp, "cp -f %s /var/cache/mondo-archive", outfile);
  run_program_and_log_output(tmp);
  sprintf (tmp, "%s/NFS-DEV", bkpinfo->tmpdir);
  write_one_liner_data_file (tmp, nfs_dev);

  sprintf (tmp, "%s/NFS-CLIENT-IPADDR", bkpinfo->tmpdir);
  write_one_liner_data_file (tmp, nfs_client_ipaddr);
  sprintf (tmp, "%s/NFS-SERVER-IPADDR", bkpinfo->tmpdir);
  write_one_liner_data_file (tmp, nfs_server_ipaddr);
  sprintf (tmp, "%s/NFS-SERVER-MOUNT", bkpinfo->tmpdir);
  write_one_liner_data_file (tmp, bkpinfo->nfs_mount);
  sprintf (tmp, "%s/NFS-SERVER-PATH", bkpinfo->tmpdir);
  write_one_liner_data_file (tmp, bkpinfo->nfs_remote_dir);
  log_it ("Finished storing NFS configuration");
}



/*************************************************************************
 * find_cdrw_device() -- Hugo Rabson                                     *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * returns:                                                              *
 *************************************************************************/
int
find_cdrw_device (char *cdrw_device)
{
	/** buffers *************************/
  /*char scratchfile[MAX_STR_LEN]; removed 19 apr 02 stan benoit */
  char comment[MAX_STR_LEN];
  char devices_file[MAX_STR_LEN];
  char tmp[MAX_STR_LEN];
  char cdr_exe[MAX_STR_LEN];
  char command[MAX_STR_LEN];

  if (!system("which dvdrecord > /dev/null 2> /dev/null"))
    { strcpy(cdr_exe, "dvdrecord"); }
  else
    { strcpy(cdr_exe, "cdrecord"); }
  sprintf(command, "%s -scanbus | tr -s '\t' ' ' | grep \"[0-9]*,[0-9]*,[0-9]*\" | grep -v \"[0-9]*) \\*\" | grep CD | cut -d' ' -f2 | head -n1", cdr_exe);
  strcpy (tmp, call_program_and_get_last_line_of_output(command));
  if (strlen (tmp) < 2)
    {
      log_to_screen ("Could not find CDRW device");
      return 1;
    }
  else
    {
      strcpy (cdrw_device, tmp);
      sprintf (comment, "Found CDRW device - %s", cdrw_device);
      log_to_screen (comment);
    }
  unlink (devices_file);
  return (0);



}




/*************************************************************************
 * help_screen() -- Hugo Rabson                                          *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * returns:                                                              *
 *************************************************************************/
void
help_screen ()
{
  fprintf (stderr, "Type 'man mondo-archive' for more information\n");
  exit (1);
}


/*************************************************************************
 * whine_if_not_found() -- Hugo Rabson                                   *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * returns:                                                              *
 *************************************************************************/
int
whine_if_not_found (char *fname)
{
	/** buffers ****/
  char command[MAX_STR_LEN];
  char errorstr[MAX_STR_LEN];


  sprintf (command, "which %s > /dev/null 2> /dev/null", fname);
  sprintf (errorstr, "Please install '%s'. I cannot find it on your system.",
	   fname);
  if (system (command))
    {
      log_to_screen (errorstr);
      log_to_screen
	("There may be hyperlink at http://www.microwerks.net/~hugo/download.html which");
      log_to_screen ("will take you to the relevant (missing) package.");
      return (1);
    }
  else
    {
      return (0);
    }
}


/*************************************************************************
 * some_basic_system_sanity_checks() -- Hugo Rabson                      *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * returns:                                                              *
 *************************************************************************/
int
some_basic_system_sanity_checks ()
{

	/** buffers *************/
  char tmp[MAX_STR_LEN];

	/** int's ****************/
  int retval = 0;



  mvaddstr_and_log_it (g_currentY, 0,
		       "Checking sanity of your Linux distribution");
  strcpy (tmp,
	  call_program_and_get_last_line_of_output
	  ("free | grep Mem | head -n1 | tr -s ' ' '\t' | cut -f2"));
  if (atol (tmp) < 35000)
    {
      retval++;
      log_to_screen ("You must have at least 32MB of RAM to use Mondo.");
    }
  if (atol (tmp) < 66000)
    {
      log_to_screen
	("WARNING! You have very little RAM. Please upgrade to 64MB or more.");
    }
  if (system ("mke2fs -V > /dev/null 2> /dev/null"))
    {
      retval++;
      log_to_screen
	("Unable to find mke2fs in system path. Please use \"su -\", not \"su\" to become root. OK? ...and please don't e-mail the mailing list about this, or me. :-) Just read the message. :)");
    }
  if (run_program_and_log_output ("cat /proc/devices | grep ramdisk"))
    {
      retval++;
      log_to_screen
	("Please recompile your kernel to include ramdisk support.");
    }
  retval += whine_if_not_found ("mke2fs");
  retval += whine_if_not_found ("afio");
  retval += whine_if_not_found ("buffer");
  retval += whine_if_not_found ("mkisofs");
  if (system("which dvdrecord > /dev/null 2> /dev/null")) { retval += whine_if_not_found ("cdrecord"); }
  retval += whine_if_not_found ("bzip2");
  retval += whine_if_not_found ("awk");
  retval += whine_if_not_found ("md5sum");
  retval += whine_if_not_found ("strings");
  retval += whine_if_not_found ("mindi");
  strcpy (tmp,
	  call_program_and_get_last_line_of_output
	  ("mount | grep \"cdr[om|w]\""));
  if (strcmp ("", tmp))
    {
      retval++;
      fatal_error ("Your CD-ROM drive is mounted. Please unmount it.");
    }
  if (!does_file_exist ("/etc/modules.conf"))
    {
      retval++;
      log_to_screen ("Please find out what happened to /etc/modules.conf");
    }
/* 
[ "`bzip2 -V | grep Version | tr -s '#' '|' | sed s/Version/#/ | cut -d'#' -f2 | sed s/,/' '/ | cut -d' ' -f2 | grep -x "0\.[8-9][0-9].*"`"
` ] && FatalError "Your bzip2 is woefully out of date (ver=$bzip_ver)" 
*/
  if (retval)
    {
      mvaddstr_and_log_it (g_currentY++, 74, "Failed.");
    }
  else
    {
      mvaddstr_and_log_it (g_currentY++, 74, "Done.");
    }
  return (retval);
}





void terminate_daemon(int sig)
{
   log_to_screen("Mondoarchive is terminating in response to a signal from the OS");
   finish(254);
}






/*************************************************************************
 * where_is_root_mounted() -- Hugo Rabson                                *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * returns:                                                              *
 *************************************************************************/
char *
where_is_root_mounted ()
{
	/** buffers *****************/
  static char tmp[MAX_STR_LEN];



  strcpy (tmp,
	  call_program_and_get_last_line_of_output
	  ("mount | grep \" on / \" | cut -d' ' -f1 | sed s/[0-9]// | sed s/[0-9]//"));
  return (tmp);
}



/*************************************************************************
 * wipe_temp_and_scratch_dirs() -- Hugo Rabson                           *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * returns:                                                              *
 *************************************************************************/
void
wipe_temp_and_scratch_dirs (struct s_bkpinfo *bkpinfo)
{

  log_to_screen
    ("wipe_temp_and_scratch_dirs(struct s_bkpinfo *bkpinfo) hasn't been written yet");
}



/*************************************************************************
 * find_home_of_exe() -- Hugo Rabson                                     *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * returns:                                                              *
 *************************************************************************/
char *
find_home_of_exe (char *fname)
{
	/** buffers **********************/
  static char output[MAX_STR_LEN];
  char incoming[MAX_STR_LEN];
  char command[MAX_STR_LEN];

  incoming[0] = '\0';
 /**********************************/

  sprintf (command, "which %s", fname);
  strcpy (incoming, call_program_and_get_last_line_of_output (command));
  if (incoming[0] != '\0')
    {
      sprintf (command, "file %s | awk '{print $NF;}'", incoming);
      strcpy (incoming, call_program_and_get_last_line_of_output (command));
    }
  if (incoming[0] != '\0')
    {
      sprintf (command, "dirname %s", incoming);
      strcpy (incoming, call_program_and_get_last_line_of_output (command));
    }
  strcpy (output, incoming);
  return (output);
}



/*************************************************************************
 * estimate_noof_media_required() -- Hugo Rabson                         *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * returns:                                                              *
 *************************************************************************/
void
estimate_noof_media_required (struct s_bkpinfo *bkpinfo)
{
	/** buffers ****************/
  char tmp[MAX_STR_LEN];

	/** longs  ****************/
  long scratchL;


  log_it ("Estimating number of media required...");
  scratchL =
    g_noof_sets * bkpinfo->optimal_set_size +
    size_of_all_biggiefiles_K (bkpinfo);
  scratchL = (scratchL / 1024) / bkpinfo->media_size[1];
  scratchL++;
  if (bkpinfo->use_lzo)
    {
      scratchL = (scratchL * 2) / 3;
    }
  else
    {
      scratchL = scratchL / 2;
    }
  if (!scratchL)
    {
      scratchL++;
    }
  if (scratchL <= 1)
    {
      sprintf (tmp,
	       "Your backup will probably occupy a single CD/tape/ISO. Maybe two.");
    }
  else
    {
      sprintf (tmp, "Your backup will occupy approximately %s media.",
	       number_to_text ((int) (scratchL + 1)));
    }
  log_to_screen (tmp);
}









void set_signals(int on)
/*
Purpose:Turn on/off signal-trapping
Params:	on - turn on or off (true=on, false=off)
Return: None
*/
{
  int signals[]= { SIGKILL, SIGPIPE, SIGTERM, SIGHUP, SIGTRAP, SIGABRT, SIGINT, SIGSTOP, 0 };
  int i;
  for (i=0; signals[i]; i++)
    {
      if (on)
        { signal(signals[i], terminate_daemon); }
      else
        { signal(signals[i], termination_in_progress); }
    }
}




/*-----------------------------------------------------------*/



void termination_in_progress(int sig)
{
  log_it("Termination in progress");
  usleep(1000);
  pthread_exit(0);
}







