/***************************************************************************
mondo-tool.c  -  description
-----------------

begin: Fri Apr 19 16:40:35 EDT 2002
copyright : (C) 2002 Mondo  Hugo Rabson
email	  : Hugo Rabson <hugo@firstlinux.net>
edited by : by Stan Benoit 4/2002
email     : troff@nakedsoul.org
***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
/* mondo-tools.c                   Hugo Rabson                     07/10/2002


07/10
- use smaller datablock in openin_tape
- added lots of =NULL's after pclose()'s for tape users

06/29
- add signal-trapping code

06/19
- re-enabled the "I won't help unless you send me your log" (or something
  like that) comments in fatal_error(); some people just don't give a rat's
  ass & expect me to wipe their behinds for free...

06/07
- don't open tape if g_tape_stream!=NULL (return 0 instead) --- openin_tape()
- closein_tape() now sets g_tape_stream to NULL at end

05/27
- pot. underflow fixed by patch from user (i>0 was missing; he added it)

04/27
- changed mojo-jojo-for-vice-prez (lockfile) to `mktemp -q /tmp/blah.XXXXXX`

04/15
- commented out some of the fatal_error() feedback

04/08
- friendlier fatal_error() message
- add manual_cd_tray to bkpinfo

03/31
- set bkpinfo's restore_path[] to "" in reset_bkpinfo()

03/03
- default compression level is now 2

02/01 - 02/28
- misc clean-ups
- 1MB buffer when reading from cdstream (was 13MB)
- disable 'checksum=' echo at end of each write to tape
- added lots of cdstream-related code
- unzip all.tar.gz to get mondo-restore.cfg, when openin'ing tape/cdstrema
- added "author can and will..." msg to fatal_error()
- added a 2nd buffer to the output (tape)
- removed '-w' from call to find size of partition
- changed Four, Five, Six, etc. to four, five, six, etc. in subroutine
  number_to_text()
- changed get_phys_size_of_drive() to read value from fdisk, not dmesg
- added size_of_all_bigfiles()
- fixed bug in size_of_all_biggiefiles_K()
- added write_cfg_var(char *label, char *value)
- added read_cfg_var(char *label, char *value)

01/01 - 01/31
- commented out, "Let's see what happens, shall we?" in two places
- don't make log_it() check last line of logfile before logging
- fixed silly bug in find_cdrom_device()
- handles files >2GB in size
- added load_filelist(), save_filelist(), free_filelist()
- disabled block-level checksums in read_file..() and write_file..()
- fixed silly bug in slice_fname()
- improved fatal_error()'s feedback
- changed g_tape_position to g_tape_posK
- I was implementing a FIFO and a fork but I changed my mind
- sleep()-related fix to eval_call()
- does_file_exist() is more sane with devices now
- improved visual feedback when writing/skipping data disks at start of tape
- added calc_checksum_list_file() stuff
- improved log_file_to_screen()
- added prototype write_file_to_tape_from_stream()
- stopped an fread() from returning an error when in fact it was just
  returning the number of bytes read
- tweaked is_this_device_mounted()
- tweaked run_program_and_log_output()
- log_it() won't log a blank (zero-length) line anymore
- fixed silly bug in last_line_of_file()
- added calc_checksum_of_file(), {write,read}_header_block_to_tape()
- adding some tape tools

12/01 - 12/30
- in text mode, don't write string to screen if it matches last line
  of log file (i.e. don't double-bounce)
- some g_text_mode-related enhancements
- replaced &> /dev/null with 2> /dev/null
- replaced '/tmp/run-prog.txt' with '/tmp/mondo-run-prog.txt'
- look at /proc/mounts, not the output of 'mount', when trying to establish
  whether a partition is mounted or not
- found several popen/pclose-type errors; fixed them
- subtract 5% from get_phys_size_of_drive()'s output
- took out all '/ /' comments
- now, log_it() only writes \n on end of line if line doesn't
  have \n on the end already
- all greps pipe 2> /dev/null or &> /dev/null, now
- fixed obscure bug in count_lines_in_file()
- don't say, "Press ENTER to quit" if finish(0) called. Just quit.
- call_program_and_log_output() doesn't put extra \n on everything
- fdisk -l 2> /dev/null --- note the '2> /dev/null' to quieten down
  does_partition_exist()
- minor code clean-ups; turned on '-Wall'

11/01 - 11/30
- moved strip_spaces() from mondo-newt.c to mondo-tools.c
- preliminary code review

Created on Sept 1st, 2001
*/



#include "my-stuff.h"
#include "mondostructures.h"
#include "mondotools.h"

/** init globals *********************************************************/

int g_currentY = 3;		/* purpose */
int g_current_media_number = 1;	/* purpose */

long g_maximum_progress = 999;	/* purpose */
long g_current_progress = -999;	/* purpose */
long g_start_time = 0;		/* purpose */

long g_original_noof_lines_in_filelist = 0;	/*purpose */

/******************** end globals ****************************************/

/** used in openin_cdstream() *******************************************/
#define BUFFER_EXE_STUFF "-m 13m -p 85"

/*************************************************************************
 * updcrc() -- Hugo Rabson                                               *
 *                                                                       *
 * Purpose:  update crc                                                  *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
unsigned int
updcrc (unsigned int crc, unsigned int c)
{
  unsigned int tmp;
  tmp = (crc >> 8) ^ c;
  crc = (crc << 8) ^ crctttab[tmp & 255];
  return crc;
}

/*************************************************************************
 * updcrcr() -- Hugo Rabson                                              *
 *                                                                       *
 * Purpose:  update crc reverse                                          *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
unsigned int
updcrcr (crc, c)
     unsigned int crc;
     unsigned int c;
{
  unsigned int tmp;
  tmp = crc ^ c;
  crc = (crc >> 8) ^ crc16tab[tmp & 0xff];
  return crc;
}


/*************************************************************************
 * calc_checksum_of_file() -- Hugo Rabson                                *
 *                                                                       *
 * Purpose:	 	                                                 *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
char *
calc_checksum_of_file (char *filename)
{
	/** buffers ******************************************************/
  static char output[MAX_STR_LEN];
  char command[MAX_STR_LEN];
  char tmp[MAX_STR_LEN];

	/** pointers *****************************************************/
  char *p;
  FILE *fin;

	/** initialize pointers ******************************************/

  p = output;

	/*****************************************************************/

  if (does_file_exist (filename))
    {
      sprintf (command, "md5sum \"%s\"", filename);
      fin = popen (command, "r");
      if (fin)
	{
	  fgets (output, MAX_STR_LEN, fin);
	  p = strchr (output, ' ');
	  pclose (fin);
	}
    }
  else
    {
      sprintf (tmp, "File '%s' not found; cannot calc checksum", filename);
      log_it (tmp);
    }
  if (p)
    {
      *p = '\0';
    }
  return (output);
}



/*************************************************************************
 * calc_file_ugly_minichecksum() -- Hugo Rabson                          *
 *                                                                       *
 * Purpose:  update crc                                                  *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
char *
calc_file_ugly_minichecksum (char *curr_fname)
{

	/** buffers ******************************************************/
  static char curr_cksum[1000];

	/** pointers *****************************************************/
  char *sz_ptr;

	/** structures ***************************************************/
  struct stat buf;

  /** initialize data ****************************************************/
  curr_cksum[0] = '\0';

	/*****************************************************************/

  if (lstat (curr_fname, &buf))
    {
      return (curr_cksum);
    }

  sprintf (curr_cksum, "%ld-%ld-%ld", (long) (buf.st_size),
	   (long) (buf.st_mtime), (long) (buf.st_ctime));
  return (curr_cksum);
  /*
     }
     CalcCksum(curr_cksum,curr_fname);
   */
  /*
     sprintf(syscall,"cksum \"%s\" 2> /dev/null",curr_fname);

     pout=popen(syscall,"r");
     if (!pout)
     {return(curr_cksum);}
     fgets(curr_cksum,999,pout);
     pclose(pout);
   */

  sz_ptr = (char *) strchr (curr_cksum, ' ');
  if (!sz_ptr)
    {
      sz_ptr = curr_cksum;
    }
  *sz_ptr = '\0';
  return (curr_cksum);
}

/*************************************************************************
 * *call_program_and_get_last_line_of_ouput() -- Hugo Rabson             *
 *                                                                       *
 * Purpose:  						                 *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/

char *
call_program_and_get_last_line_of_output (char *call)
{
	/** buffers ******************************************************/
  static char result[MAX_STR_LEN];
  char tmp[MAX_STR_LEN];

	/** pointers *****************************************************/
  FILE *fin;

	/** initialize data **********************************************/
  result[0] = '\0';

  /***********************************************************************/
  if ((fin = popen (call, "r")))
    {
      for (fgets (tmp, MAX_STR_LEN, fin); !feof (fin);
	   fgets (tmp, MAX_STR_LEN, fin))
	{
	  if (strlen (tmp) > 1)
	    {
	      strcpy (result, tmp);
	    }
	}
      pclose (fin);
    }
  strip_spaces (result);
  return (result);
}



/*************************************************************************
 * center_string() -- Hugo Rabson                                        *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
void
center_string (char *in_out, int width)
{
	/** buffers ******************************************************/
  char scratch[MAX_STR_LEN];

	/** pointers *****************************************************/
  char *p;

	/** int's ********************************************************/
  int i;			/* purpose */
  int len;			/* purpose */
  int mid;			/* purpose */
  int x;			/* purpose */

	/*****************************************************************/

  if (strlen (in_out) == 0)
    {
      return;
    }
  for (p = in_out; *p == ' '; p++);
  strcpy (scratch, p);
  len = strlen (scratch);
  mid = width / 2;
  x = mid - len / 2;
  for (i = 0; i < x; i++)
    {
      in_out[i] = ' ';
    }
  in_out[i] = '\0';
  strcat (in_out, scratch);
}




/*************************************************************************
 * closein_cdstream() -- Hugo Rabson                                     *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
int
closein_cdstream (struct s_bkpinfo *bkpinfo)
{
  return (closein_tape (bkpinfo));
}



/*************************************************************************
 * closein_tape() -- Hugo Rabson                                         *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
int
closein_tape (struct s_bkpinfo *bkpinfo)
{
	/** int's ********************************************************/
  int retval = 0;
  int res = 0;
  int ctrl_chr = '\0';

	/** buffers ******************************************************/
  char fname[MAX_STR_LEN];

	/** long long's **************************************************/
  long long size;


  res = read_header_block_from_tape (&size, fname, &ctrl_chr);
  retval += res;
  if (ctrl_chr != BLK_END_OF_BACKUP)
    {
      wrong_marker (BLK_END_OF_BACKUP, ctrl_chr);
    }
  res = read_header_block_from_tape (&size, fname, &ctrl_chr);
  retval += res;
  if (ctrl_chr != BLK_END_OF_TAPE)
    {
      wrong_marker (BLK_END_OF_TAPE, ctrl_chr);
    }
  pclose (g_tape_stream);
  g_tape_stream = NULL;
  return (retval);
}


/*************************************************************************
 * count_lines_in_file() -- Hugo Rabson                                  *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
long
count_lines_in_file (char *filename)
{

	/** buffers ******************************************************/
  char command[MAX_STR_LEN];
  char incoming[MAX_STR_LEN];
  char tmp[MAX_STR_LEN];

	/** long *********************************************************/
  long noof_lines = -1L;

	/** pointers *****************************************************/
  FILE *fin;

  /** initialize [0] to null *********************************************/
  incoming[0] = '\0';

/*
  sprintf(tmp,"Counting the lines in file '%s'",filename);
  log_it(tmp);
*/
  if (!does_file_exist (filename))
    {
      sprintf (tmp,
	       "%s does not exist, so I cannot found the number of lines in it",
	       filename);
      log_it (tmp);
      return (0);
    }
  sprintf (command, "grep -n \"\" %s | tail -n1 | cut -d':' -f1 2> /dev/null",
	   filename);
  fin = popen (command, "r");
  if (fin)
    {
      if (feof (fin))
	{
	  noof_lines = 0;
	}
      else
	{
	  fgets (incoming, MAX_STR_LEN - 1, fin);
	  while (strlen (incoming) > 0
		 && incoming[strlen (incoming) - 1] < 32)
	    {
	      incoming[strlen (incoming) - 1] = '\0';
	    }
	  noof_lines = atol (incoming);
	}
      pclose (fin);
    }
/*
  sprintf(tmp,"File '%s' has %ld lines in it",filename,noof_lines);
  log_it(tmp);
*/
  return (noof_lines);
}



/*************************************************************************
 * does_file_exist() -- Hugo Rabson                                      *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
bool
does_file_exist (char *filename)
{

	/** structures ***************************************************/
  struct stat buf;

	/*****************************************************************/

  if (lstat (filename, &buf))
    {
      return (FALSE);
    }
  else
    {
      return (TRUE);
    }
}





/*************************************************************************
 * does_partition_exist() -- Hugo Rabson                                 *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
int
does_partition_exist (char *drive, int partno)
{
	/** buffers *****************************************************/
  char program[MAX_STR_LEN];
  char incoming[MAX_STR_LEN];
  char searchstr[MAX_STR_LEN];

	/** ints ********************************************************/
  int res = 0;

	/** pointers ****************************************************/
  FILE *fin;


	/****************************************************************/
  sprintf (searchstr, "%s%d ", drive, partno);
  sprintf (program, "fdisk -l %s 2> /dev/null", drive);
  fin = popen (program, "r");
  if (!fin)
    {
      return (0);
    }
  fgets (incoming, MAX_STR_LEN - 1, fin);
  for (res = 0; !feof (fin) && !res; fgets (incoming, MAX_STR_LEN - 1, fin))
    {
      if (strstr (incoming, searchstr))
	{
	  res = 1;
	}
    }
  pclose (fin);
  return (res);
}




/*************************************************************************
 * does_string_exist_in_boot_block() -- Hugo Rabson                      *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
bool
does_string_exist_in_boot_block (char *dev, char *str)
{
	/** buffers *****************************************************/
  char command[MAX_STR_LEN];

	/****************************************************************/

  sprintf (command,
	   "dd if=%s bs=446 count=1 2> /dev/null | strings | grep \"%s\" > /dev/null 2> /dev/null",
	   dev, str);
  if (system (command))
    {
      return (FALSE);
    }
  else
    {
      return (TRUE);
    }
}





/*************************************************************************
 * exclude_nonexistent_files() -- Hugo Rabson                            *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
void
exclude_nonexistent_files (char *inout)
{
  char infname[MAX_STR_LEN];
  char outfname[MAX_STR_LEN];
  char tmp[MAX_STR_LEN];
  char incoming[MAX_STR_LEN];

	/** int **********************************************************/
  int i;

	/** pointers *****************************************************/
  FILE *fin, *fout;


 /** end vars ************************************************************/

  sprintf (infname, "%s.in", inout);
  sprintf (outfname, "%s", inout);
  sprintf (tmp, "cp -f %s %s", inout, infname);
  run_program_and_log_output (tmp);
  fin = fopen (infname, "r");
  fout = fopen (outfname, "w");
  for (fgets (incoming, MAX_STR_LEN, fin); !feof (fin);
       fgets (incoming, MAX_STR_LEN, fin))
    {
      i = strlen (incoming) - 1;
      if (i >= 0 && incoming[i] < 32)
	{
	  incoming[i] = '\0';
	}
      if (does_file_exist (incoming))
	{
	  fprintf (fout, "%s\n", incoming);
	}
      else
	{
	  sprintf (tmp, "Excluding '%s'-nonexistent\n", incoming);
	  log_it (tmp);
	}
    }
  fclose (fout);
  fclose (fin);
  unlink (infname);
}





/*************************************************************************
 * fatal_error() -- Hugo Rabson                                          *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
void
fatal_error (char *error_string)
{
	/** buffers ******************************************************/
  char fatalstr[MAX_STR_LEN] = "-------FATAL ERROR---------";

	/** end vars *****************************************************/

  if (!g_text_mode)
    {
      log_it (fatalstr);
      log_it (error_string);
      newtFinished ();
    }
  printf ("%s\n", fatalstr);
  printf ("%s\n", error_string);
  if (strstr (error_string, "Mindi"))
    {
      printf ("Last 6 lines of mindi.log:\n");
      system ("tail /var/log/mindi.log");
    }
  system("cat /var/log/mondo-archive.log | gzip -9 > /tmp/MA.log.gz 2> /dev/null");
  printf("Author can and will render assistance only to users who include a gzipped copy\n");
  printf("of /var/log/mondo-archive.log and /var/log/mindi.log in their bug report.\n");
  printf("(Or /tmp/mondo-restore.log if it is a restore-time problem)\n");
  printf
    ("Please read the FAQ and the logs before you e-mail the mailing list.\n");
  if (does_file_exist("/tmp/MA.log.gz"))
    { printf("FYI, I have gzipped the log and saved it to /tmp/MA.log.gz"); }
  printf ("Mondo has aborted.\n");
  exit (1);
}


/*************************************************************************
 * find_and_mount_actual_cd() -- Hugo Rabson                             *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
int
find_and_mount_actual_cd (struct s_bkpinfo *bkpinfo, char *mountpoint)
{
	/** buffers ******************************************************/
  char tmp[MAX_STR_LEN];

	/** int's  *******************************************************/
  int res;

	/** end vars *****************************************************/

  mvaddstr_and_log_it (g_currentY, 0, "");
/*  if (bkpinfo->media_device[0]=='\0')
    {
*/
  res = find_cdrom_device (bkpinfo->media_device);
  sprintf (tmp, "g_cdrom_device = %s", bkpinfo->media_device);
  log_it (tmp);
  if (res)
    {
      if (!popup_and_get_string
	  ("CD-RW device", "Please enter your CD-RW's /dev device",
	   bkpinfo->media_device))
	{
	  res = 1;
	}
      else
	{
	  res = 0;
	}
    }
//    }
  if (res)
    {
      return (res);
    }
  if (mount_CDROM_here (bkpinfo->media_device, mountpoint))
    {
      strcpy (bkpinfo->media_device, "/dev/scd0");
      if ((res = mount_CDROM_here (bkpinfo->media_device, mountpoint)))
	{
	  strcpy (bkpinfo->media_device, "/dev/sr0");
	  if ((res = mount_CDROM_here (bkpinfo->media_device, mountpoint)))
	    {
	      strcpy (bkpinfo->media_device, "/dev/scd1");
	      if ((res =
		   mount_CDROM_here (bkpinfo->media_device, mountpoint)))
		{
		  strcpy (bkpinfo->media_device, "/dev/sr1");
		  if ((res =
		       mount_CDROM_here (bkpinfo->media_device, mountpoint)))
		    {
		      if (!popup_and_get_string
			  ("CDRW device",
			   "Please enter your CDRW's /dev device",
			   bkpinfo->media_device))
			{
			  res = 1;
			}
		      else
			{
			  res =
			    mount_CDROM_here (bkpinfo->media_device,
					      mountpoint);
			}
		    }
		}
	    }
	}
    }
  return (res);
}



/*************************************************************************
 * find_cdrom_device() -- Hugo Rabson                                    *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
int
find_cdrom_device (char *output)
{
	/** pointers *****************************************************/
  FILE *fin;
  char *p;
  char *q;
  char *r;

	/** bool's *******************************************************/
  bool found_it = FALSE;

	/** buffers ******************************************************/
  char tmp[MAX_STR_LEN];
  char phrase_one[MAX_STR_LEN];
  char phrase_two[MAX_STR_LEN];
  char command[MAX_STR_LEN];

	/** intialize ****************************************************/
  output[0] = '\0';
  phrase_one[0] = '\0';
  phrase_two[0] = '\0';

	/** end vars *****************************************************/

  fin = popen ("cdrecord -scanbus", "r");
  if (!fin)
    {
      return (1);
    }
  for (fgets (tmp, MAX_STR_LEN, fin); !feof (fin);
       fgets (tmp, MAX_STR_LEN, fin))
    {
      p = strchr (tmp, '\'');
      if (p)
	{
	  q = strchr (++p, '\'');
	  if (q)
	    {
	      for (r = q; *(r - 1) == ' '; r--);
	      *r = '\0';
	      strcpy (phrase_one, p);
	      p = strchr (++q, '\'');
	      if (p)
		{
		  q = strchr (++p, '\'');
		  if (q)
		    {
		      while (*(q - 1) == ' ')
			{
			  q--;
			}
		      *q = '\0';
		      strcpy (phrase_two, p);
		    }
		}
	    }
	}
    }
  pclose (fin);
  sprintf (command, "dmesg | grep %s", phrase_two);
  fin = popen (command, "r");
  if (fin)
    {
      for (fgets (tmp, MAX_STR_LEN, fin); !feof (fin);
	   fgets (tmp, MAX_STR_LEN, fin))
	{
	  if (tmp[0] != ' ' && tmp[1] != ' ')
	    {
	      p = strchr (tmp, ':');
	      if (p)
		{
		  *p = '\0';
		  sprintf (output, "/dev/%s", tmp);
		  found_it = TRUE;
		}
	    }
	}
      pclose (fin);
    }
  if (found_it)
    {
      return (0);
    }
  else
    {
      return (1);
    }
}


/*************************************************************************
 * finish() -- Hugo Rabson                                               *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
void
finish (int signal)
{
//  char tmp[MAX_STR_LEN];

  /*  if (signal==0) { popup_and_OK("Please press <enter> to quit."); } */

  /* newtPopHelpLine(); */

  if (!g_text_mode)
    {
      newtFinished ();
    }
  /* system("clear"); */
  if (does_file_exist ("/tmp/mondo-restore.log"))
    {
      printf ("Type 'less /tmp/mondo-restore.log' to see the output log\n");
    }
  exit (signal);
}



/*************************************************************************
 * free_filelist() -- Hugo Rabson                                        *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
void
free_filelist (struct s_node *filelist)
{
	/** int's ********************************************************/
  static int depth = 0;
  int percentage;

	/** long's *******************************************************/
  static long i = 0;

	/** end vars *****************************************************/

  if (depth == 0)
    {
      open_evalcall_form ("Freeing memory");
      log_to_screen ("Freeing memory formerly occupied by filelist");
    }
  depth++;

  if (filelist->ch == '\0')
    {
      if (!(i++ % 1111))
	{
	  percentage = (int) (i * 100 / g_original_noof_lines_in_filelist);
	  update_evalcall_form (percentage);

	}
    }

  if (filelist->right)
    {
      free_filelist (filelist->right);
      filelist->right = NULL;
    }
  if (filelist->down)
    {
/*      if (!(i++ %39999)) { update_evalcall_form(0); } */
      free_filelist (filelist->down);
      filelist->down = NULL;
    }
  filelist->ch = '\0';
  free (filelist);
  depth--;
  if (depth == 0)
    {
      close_evalcall_form ();
      log_it ("Finished freeing memory");
    }
}



/*************************************************************************
 * friendly_sizestr_to_sizelong() -- Hugo Rabson                         *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
long
friendly_sizestr_to_sizelong (char *incoming)
{
	/** long's ******************************************************/
  long outval;

	/** int's  ******************************************************/
  int i;

	/** buffers *****************************************************/
  char tmp[MAX_STR_LEN];

	/** char's ******************************************************/
  char ch;

  /** end vars **********************************************************/


  strcpy (tmp, incoming);
  for (i = 0; i < strlen (tmp) && isdigit (tmp[i]); i++);
  ch = tmp[i];
  tmp[i] = '\0';
  outval = atol (tmp);
  if (ch == 'g')
    {
      outval = outval * 1024;
    }
  else if (ch == 'k')
    {
      outval = outval / 1024;
    }
  return (outval);
}




/*************************************************************************
 * get_last_filelist_number() -- Hugo Rabson                             *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
int
get_last_filelist_number (struct s_bkpinfo *bkpinfo)
{
	/** buffers ******************************************************/
  char val_sz[MAX_STR_LEN];
  char cfg_fname[MAX_STR_LEN];
/*  char tmp[MAX_STR_LEN]; remove stan benoit apr 2002 */

	/** long *********************************************************/
  long val_i;

	/** end vars *****************************************************/

  sprintf (cfg_fname, "%s/mondo-restore.cfg", bkpinfo->tmpdir);
  read_cfg_var (cfg_fname, "last-filelist-number", val_sz);
  val_i = atoi (val_sz);
  if (val_i <= 0)
    {
      val_i = 500;
    }
  return (val_i);

/*
  FILE*fin;
  char fname[MAX_STR_LEN],incoming[MAX_STR_LEN+1];
  int res=-1,i;
  sprintf(fname, "%s/LAST-FILELIST-NUMBER",bkpinfo->tmpdir);
  fin=fopen(fname,"r");
  if (fin)
    {
      fgets(incoming,MAX_STR_LEN,fin);
      fclose(fin);
      for(i=strlen(incoming); incoming[i-1]<32; i--);
      incoming[i]='\0';
      res=atoi(incoming);
    }
  else
    {
      log_it("WARNING - unable to find LAST-FILELIST-NUMBER");
    }
  return(res);
*/
}




/*************************************************************************
 * get_phys_size_of_drive() -- Hugo Rabson                               *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
long
get_phys_size_of_drive (char *drive)
{
	/** pointers *****************************************************/
  FILE *fin;
  char *p;
  char *q;
  char *r;

	/** buffers ******************************************************/
  char tmp[MAX_STR_LEN];
  char command[MAX_STR_LEN];

	/** long *********************************************************/
  long outL;
  long tempLa;
  long tempLb;
  long tempLc;

	/** end vars *****************************************************/

  sprintf (command,
	   "fdisk -l %s | head -n4 | tr -s '\n' '\t' | tr -s ' ' '\t' | cut -f8,14,16",
	   drive);
  strcpy (tmp, call_program_and_get_last_line_of_output (command));
  if (tmp[0])
    {
      p = tmp;
      q = strchr (p, ' ');
      if (q)
	{
	  *(q++) = '\0';
	  r = strchr (q, ' ');
	  if (r)
	    {
	      *(r++) = '\0';
	      tempLa = atol (p);
	      tempLb = atol (q);
	      tempLc = atol (r);
	      outL = tempLa * tempLb / 1024 * tempLc / 1024;
	      if (outL > 100)
		{
		  return (outL);
		}
	    }
	}
    }
  /* else, do it the old-fashioned way */
  p = strrchr (drive, (int) '/');
  if (p)
    {
      strcpy (tmp, p + 1);
    }
  else
    {
      return (-1);
    }
  sprintf (command, "dmesg | grep %s 2> /dev/null", tmp);
  fin = popen (command, "r");
  fgets (tmp, MAX_STR_LEN - 1, fin);
  while (!feof (fin) && !strstr (tmp, "GB") && !strstr (tmp, "MB"))
    {
      fgets (tmp, MAX_STR_LEN - 1, fin);
    }
  pclose (fin);
  if (!(p = strstr (tmp, "GB")) && !(p = strstr (tmp, "MB")))
    {
      sprintf (tmp, "Cannot find %s's size: dmesg isn't helping either.",
	       drive);
      log_it (tmp);
      return (-1);
    }
  for (; !isdigit (*(p - 1)); p--);
  *p = '\0';
  for (p--; isdigit (*(p - 1)); p--);
  outL = atol (p);
  if (outL <= 0)
    {
      return (-1);
    }
  if (strstr (tmp, "GB"))
    {
      outL = outL * 1024;
    }
  return (outL * 20 / 19);
}



/*************************************************************************
 * get_time() -- Hugo Rabson 					         *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
long
get_time ()
{

	/** pointers *****************************************************/
  FILE *fin;

	/** buffers ******************************************************/
  char incoming[MAX_STR_LEN];

	/** end vars *****************************************************/

  fin = popen ("date +%s", "r");
  fgets (incoming, MAX_STR_LEN - 1, fin);
  pclose (fin);
  return (atol (incoming));
}


/*************************************************************************
 * grab_percentage_from_last_line_of_file() -- Hugo Rabson               *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
int
grab_percentage_from_last_line_of_file (char *filename)
{

	/** buffers ******************************************************/
  char tmp[MAX_STR_LEN];
  char lastline[MAX_STR_LEN];

	/** pointers *****************************************************/
  char *p;

	/** int's ********************************************************/
  int i;


  /* strcpy(lastline,last_line_of_file(filename)); */
  strcpy (lastline, err_log_lines[NOOF_ERR_LINES - 1]);
  sprintf (tmp, "lastline = %s", lastline);
  p = strstr (lastline, "% done");
  if (!p)
    {
      p = strstr (lastline, "%% Done");
    }
  if (!p)
    {
      p = strstr (lastline, "%% ");
    }
  if (!p)
    {
      return (0);
    }
  *p = '\0';
  /*
     sprintf(tmp,"lastline = '%s'");
     log_to_screen(tmp);
   */
  for (p -= 1; *p != ' ' && p != lastline; p--);
  if (p != lastline)
    {
      p++;
    }
  i = atoi (p);
  /*
     sprintf(tmp,"'%s' --> %d",p,i);
     log_to_screen(tmp);
   */
  return (i);
}

/*************************************************************************
 * initialize_raidrec() -- Hugo Rabson                                   *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
void
initialize_raidrec (struct raid_device_record *raidrec)
{
  raidrec->raid_device[0] = '\0';
  raidrec->raid_level = 0;
  raidrec->chunk_size = 4;
  raidrec->persistent_superblock = 1;
  raidrec->data_disks.entries = 0;
  raidrec->spare_disks.entries = 0;
  raidrec->parity_disks.entries = 0;
  raidrec->failed_disks.entries = 0;
  raidrec->additional_vars.entries = 0;
}




/*************************************************************************
 * is_this_device_mounted() -- Hugo Rabson                               *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
int
is_this_device_mounted (char *device_raw)
{

	/** pointers *****************************************************/
  FILE *fin;

	/** buffers ******************************************************/
  char incoming[MAX_STR_LEN];
  char device_with_tab[MAX_STR_LEN];
  char device_with_space[MAX_STR_LEN];
  char tmp[MAX_STR_LEN];

	/** end vars *****************************************************/

  sprintf (device_with_tab, "%s\t", device_raw);
  sprintf (device_with_space, "%s ", device_raw);
  fin = fopen ("/proc/mounts", "r");
  for (fgets (incoming, MAX_STR_LEN - 1, fin); !feof (fin);
       fgets (incoming, MAX_STR_LEN - 1, fin))
    {
      if (strstr (incoming, device_with_space)
	  || strstr (incoming, device_with_tab))
	{
	  fclose (fin);
	  return (1);
	}
    }
  fclose (fin);
  sprintf (tmp, "cat /proc/swaps | grep \"%s\" > /dev/null 2> /dev/null",
	   device_with_space);
  if (!system (tmp))
    {
      return (1);
    }
  return (0);
}





/*************************************************************************
 * last_line_of_file() -- Hugo Rabson                                    *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
char *
last_line_of_file (char *filename)
{
	/** buffers ******************************************************/
  static char output[MAX_STR_LEN];
  static char command[MAX_STR_LEN];
  static char tmp[MAX_STR_LEN];

	/** pointers *****************************************************/
  FILE *fin;

	/** end vars *****************************************************/

  if (!does_file_exist (filename))
    {
      sprintf (tmp, "Tring to get last line of nonexistent file (%s)",
	       filename);
      log_it (tmp);
      output[0] = '\0';
      return (output);
    }
  sprintf (command, "cat %s | tail -n1", filename);
  fin = popen (command, "r");
  fgets (output, MAX_STR_LEN, fin);
  pclose (fin);
  while (strlen (output) > 0 && output[strlen (output) - 1] < 32)
    {
      output[strlen (output) - 1] = '\0';
    }
  return (output);
}


/*************************************************************************
 * leftpad_string() -- Hugo Rabson                                       *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
char *
leftpad_string (char *incoming, int width)
{
	/** buffers ******************************************************/
  static char output[MAX_STR_LEN];

	/** ints *********************************************************/
  int i;

	/** end vars *****************************************************/
  strcpy (output, incoming);
  for (i = strlen (output); i < width; i++)
    {
      output[i] = ' ';
    }
  output[i] = '\0';
  return (output);
}




/*************************************************************************
 * length_of_file() -- Hugo Rabson                                       *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
long long
length_of_file (char *filename)
{
	/** pointers ****************************************************/
  FILE *fin;

	/** long long **************************************************/
  long long length;

  fin = fopen (filename, "r");
  if (!fin)
    {
      return (-1);
    }
  fseek (fin, 0, SEEK_END);
  length = ftell (fin);
  fclose (fin);
  return (length);
}




/*************************************************************************
 * add_string_at_node() -- Hugo Rabson                                   *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
int
add_string_at_node (struct s_node *startnode, char *string_to_add)
{
	/** int *********************************************************/
  int noof_chars;
  int i;

	/** sturctures **************************************************/
  struct s_node *node, *newnode;

	/** char  *******************************************************/
  char char_to_add;

	/** bools *******************************************************/
  const bool sosodef = FALSE;


  noof_chars = strlen (string_to_add) + 1;	/* we include the '\0' */

/* walk across tree */
  node = startnode;
  char_to_add = string_to_add[0];
  for (node = startnode; node->right != NULL && node->ch < char_to_add;
       node = node->right);

/* walk down tree if appropriate */
  if (node->ch == char_to_add)
    {
      if (noof_chars == 1 && char_to_add == '\0')
	{
	  log_it ("Already present. ");
	  return (1);
	}
      if (!node->down)
	{
	  log_it ("Partially listed but it carries on and I don't");
	  return (1);
	}
      return (add_string_at_node (node->down, string_to_add + 1));
    }

/* else, add node(s) here */
  if (!(newnode = malloc (sizeof (struct s_node))))
    {
      log_to_screen ("failed to malloc");
      return (1);
    }

  memcpy (newnode, node, sizeof (struct s_node));
  node->right = newnode;
  node->ch = char_to_add;
  node->down = NULL;
  node->expanded = node->selected = FALSE;

  if (char_to_add == '\0')
    {
      node->selected = sosodef;
      return (0);
    }
  for (i = 1; i < noof_chars; i++)
    {
      if (!(node->down = malloc (sizeof (struct s_node))))
	{
	  log_to_screen ("failed to malloc");
	  return (1);
	}
      node = node->down;
      char_to_add = string_to_add[i];
      node->ch = char_to_add;
      node->right = node->down = NULL;
      node->expanded = node->selected = FALSE;
    }
  if (char_to_add == '\0')
    {
      node->selected = sosodef;
      return (0);
    }
  return (0);
}





/*************************************************************************
 * load_filelist() -- Hugo Rabson                                        *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
struct s_node *
load_filelist (char *filelist_fname)
{

	/** structures **************************************************/
  struct s_node *filelist;

	/** pointers ****************************************************/
  FILE *fin;

	/** buffers *****************************************************/
  char fname[MAX_STR_LEN];
  char tmp[MAX_STR_LEN];

	/** int *********************************************************/
  int percentage;

	/** long ********************************************************/
  long lines_in_filelist;
  long lino = 0;
	/** end vars ****************************************************/

  log_to_screen ("Loading filelist");
  lines_in_filelist = count_lines_in_file (filelist_fname);
  g_original_noof_lines_in_filelist = lines_in_filelist;
  if (!(filelist = malloc (sizeof (struct s_node))))
    {
      return (NULL);
    }
  filelist->ch = '\0';
  filelist->right = filelist->down = NULL;
  filelist->expanded = filelist->selected = FALSE;
  if (!(fin = fopen (filelist_fname, "r")))
    {
      return (NULL);
    }
  open_evalcall_form ("Loading filelist from disk");
  for (fgets (fname, MAX_STR_LEN, fin); !feof (fin);
       fgets (fname, MAX_STR_LEN, fin))
    {
      strip_spaces (fname);
      if (!strlen (fname))
	{
	  continue;
	}
      if (add_string_at_node (filelist, fname))
	{
	  sprintf (tmp, "Did not load '%s' into filelist", fname);
	  log_it (tmp);
	}
      if (!(++lino % 1111))
	{
	  percentage = (int) (lino * 100 / lines_in_filelist);
	  update_evalcall_form (percentage);
	}
    }
  fclose (fin);
  close_evalcall_form ();
  log_it ("Finished loading filelist");
  return (filelist);
}

/*************************************************************************
 * log_trace() --Stan Benoit                                             *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
void
log_trace (char *o)
{
	/** pointers *****************************************************/
  FILE *fout;

	/** buffers ******************************************************/
  char output[MAX_STR_LEN];

	/** int    *******************************************************/
  int i;

	/** end vars ****************************************************/


  if (o[0] == '\0')
    {
      return;
    }
  strcpy (output, o);
  i = strlen (output);
  if (i <= 0)
    {
      return;
    }
  if (output[i - 1] < 32)
    {
      output[i - 1] = '\0';
    }
  if (g_text_mode /* && !strstr(last_line_of_file(MONDO_LOGFILE),output) */ )
    {
      printf ("%s\n", output);
    }

  fout = fopen (MONDO_TRACEFILE, "a");
  if (fout)
    {
      fprintf (fout, "%s\n", output);
      fclose (fout);
    }
}



/*************************************************************************
 * log_it() -- Hugo Rabson                                               *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
void
log_it (char *o)
{
	/** pointers *****************************************************/
  FILE *fout;

	/** buffers ******************************************************/
  char output[MAX_STR_LEN];

	/** int    *******************************************************/
  int i;

	/** end vars ****************************************************/


  if (o[0] == '\0')
    {
      return;
    }
  strcpy (output, o);
  i = strlen (output);
  if (i <= 0)
    {
      return;
    }
  if (output[i - 1] < 32)
    {
      output[i - 1] = '\0';
    }
  if (g_text_mode /* && !strstr(last_line_of_file(MONDO_LOGFILE),output) */ )
    {
      printf ("%s\n", output);
    }
  fout = fopen (MONDO_LOGFILE, "a");
  if (fout)
    {
      fprintf (fout, "%s\n", output);
      fclose (fout);
    }
}



/*************************************************************************
 * log_tape_pos() -- Hugo Rabson                                         *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
void
log_tape_pos (void)
{
	/** buffers ******************************************************/
  char tmp[MAX_STR_LEN];


	/** end vars ****************************************************/

  sprintf (tmp, "Tape position -- %ld KB (%ld MB)", (long) g_tape_posK,
	   (long) g_tape_posK >> 10);
  log_it (tmp);
}







/*************************************************************************
 * make_checksum_list_file() -- Hugo Rabson                              *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
int
make_checksum_list_file (char *filelist, char *cksumlist, char *comppath)
{
	/** pointers *****************************************************/
  FILE *fin;
  FILE *fout;

	/** int   ********************************************************/
  int percentage;
  int i;
  int counter = 0;

	/** buffer *******************************************************/
  char stub_fname[1000];
  char curr_fname[1000];
  char curr_cksum[1000];
  char tmp[1000];

	/** end vars ****************************************************/
  long filelist_length;
  long curr_pos;
  long start_time;
  long current_time;
  long time_taken;
  long time_remaining;

	/** end vars ****************************************************/

  start_time = get_time ();
  filelist_length = length_of_file (filelist);
  sprintf (tmp, "filelist = %s; cksumlist = %s", filelist, cksumlist);
  log_it (tmp);
  fin = fopen (filelist, "r");
  if (fin == NULL)
    {
      log_to_screen ("Can't open filelist");
      return (1);
    }
  fout = fopen (cksumlist, "w");
  if (fout == NULL)
    {
      fclose (fin);
      log_to_screen ("Can't open checksum list");
      return (1);
    }
  for (fgets (stub_fname, 999, fin); !feof (fin);
       fgets (stub_fname, 999, fin))
    {
      if (stub_fname[(i = strlen (stub_fname) - 1)] < 32)
	{
	  stub_fname[i] = '\0';
	}
      sprintf (tmp, "%s%s", comppath, stub_fname);
      strcpy (curr_fname, tmp + 1);
      strcpy (curr_cksum, calc_file_ugly_minichecksum (curr_fname));
      fprintf (fout, "%s\t%s\n", curr_fname, curr_cksum);
      if (counter++ > 12)
	{
	  current_time = get_time ();
	  counter = 0;
	  curr_fname[37] = '\0';
	  curr_pos = ftell (fin) / 1024;
	  percentage = (int) (curr_pos * 100 / filelist_length);
	  time_taken = current_time - start_time;
	  if (percentage == 0)
	    {
	      /*              printf("%0d%% done      \r",percentage); */
	    }
	  else
	    {
	      time_remaining =
		time_taken * 100 / (long) (percentage) - time_taken;
	      sprintf (tmp,
		       "%02d%% done   %02d:%02d taken   %02d:%02d remaining  %-37s\r",
		       percentage, (int) (time_taken / 60),
		       (int) (time_taken % 60), (int) (time_remaining / 60),
		       (int) (time_remaining % 60), curr_fname);
	      log_to_screen (tmp);
	    }
	  sync ();
	}
    }
  fclose (fout);
  fclose (fin);
  log_it ("Done.");
  return (0);
}


/*************************************************************************
 * make_hole_for_file() -- Hugo Rabson                                   *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
int
make_hole_for_file (char *outfile_fname)
{
	/** buffer *******************************************************/
  char command[MAX_STR_LEN];

	/** int  *********************************************************/
  int res = 0;

	/** end vars ****************************************************/

  sprintf (command, "mkdir -p \"%s\" 2> /dev/null", outfile_fname);
  res += system (command);
  sprintf (command, "rmdir \"%s\" 2> /dev/null", outfile_fname);
  res += system (command);
  sprintf (command, "rm -f \"%s\" 2> /dev/null", outfile_fname);
  res += system (command);
  unlink (outfile_fname);
  return (0);
}




/*************************************************************************
 * maker_to_string () -- Hugo Rabson                                     *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
char *
marker_to_string (int marker)
{
	/** buffer *******************************************************/
  static char outstr[MAX_STR_LEN];


	/** end vars ****************************************************/

  switch (marker)
    {
    case BLK_START_OF_BACKUP:
      strcpy (outstr, "BLK_START_OF_BACKUP");
      break;
    case BLK_START_OF_TAPE:
      strcpy (outstr, "BLK_START_OF_TAPE");
      break;
    case BLK_START_AN_AFIO_OR_SLICE:
      strcpy (outstr, "BLK_START_AN_AFIO_OR_SLICE");
      break;
    case BLK_STOP_AN_AFIO_OR_SLICE:
      strcpy (outstr, "BLK_STOP_AN_AFIO_OR_SLICE");
      break;
    case BLK_START_AFIOBALLS:
      strcpy (outstr, "BLK_START_AFIOBALLS");
      break;
    case BLK_STOP_AFIOBALLS:
      strcpy (outstr, "BLK_STOP_AFIOBALLS");
      break;
    case BLK_STOP_BIGGIEFILES:
      strcpy (outstr, "BLK_STOP_BIGGIEFILES");
      break;
    case BLK_START_A_BIGGIE:
      strcpy (outstr, "BLK_START_A_BIGGIE");
      break;
    case BLK_START_BIGGIEFILES:
      strcpy (outstr, "BLK_START_BIGGIEFILES");
      break;
    case BLK_STOP_A_BIGGIE:
      strcpy (outstr, "BLK_STOP_A_BIGGIE");
      break;
    case BLK_END_OF_TAPE:
      strcpy (outstr, "BLK_END_OF_TAPE");
      break;
    case BLK_END_OF_BACKUP:
      strcpy (outstr, "BLK_END_OF_BACKUP");
      break;
    case BLK_ABORTED_BACKUP:
      strcpy (outstr, "BLK_ABORTED_BACKUP");
      break;
    case BLK_START_FILE:
      strcpy (outstr, "BLK_START_FILE");
      break;
    case BLK_STOP_FILE:
      strcpy (outstr, "BLK_STOP_FILE");
      break;
    default:
      sprintf (outstr, "BLK_UNKNOWN (%d)", marker);
      break;
    }
  return (outstr);
}



/*************************************************************************
 * mount_CDROM_here() -- Hugo Rabson                                     *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
int
mount_CDROM_here (char *device, char *mountpoint)
{
	/** buffer *******************************************************/
  char command[MAX_STR_LEN];



	/** end vars ****************************************************/
  sprintf (command, "mount %s -o ro -t iso9660 %s 2> /dev/null", device,
	   mountpoint);
  log_it (command);
  return (system (command));
}





/*************************************************************************
 * noof_lines_that_match_wildcard() -- Hugo Rabson                       *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
long
noof_lines_that_match_wildcard (char *filelist_fname, char *wildcard)
{
	/** long ********************************************************/
  long matches = 0;

	/** pointers ****************************************************/
  FILE *fin;

	/** buffers *****************************************************/
  char incoming[MAX_STR_LEN];

  fin = fopen (filelist_fname, "r");

	/** end vars ****************************************************/
  if (!fin)
    {
      return (0);
    }
  fgets (incoming, MAX_STR_LEN - 1, fin);
  while (!feof (fin))
    {
      if (strstr (incoming, wildcard))
	{
	  matches++;
	}
      fgets (incoming, MAX_STR_LEN - 1, fin);
    }
  fclose (fin);
  return (matches);
}





/*************************************************************************
 * number_to_text() -- Hugo Rabson                                       *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
char *
number_to_text (int i)
{

	/** buffers ******************************************************/
  static char output[MAX_STR_LEN];


	/** end vars ****************************************************/

  switch (i)
    {
    case 0:
      strcpy (output, "zero");
      break;
    case 1:
      strcpy (output, "one");
      break;
    case 2:
      strcpy (output, "two");
      break;
    case 3:
      strcpy (output, "three");
      break;
    case 4:
      strcpy (output, "four");
      break;
    case 5:
      strcpy (output, "five");
      break;
    case 6:
      strcpy (output, "six");
      break;
    case 7:
      strcpy (output, "seven");
      break;
    case 8:
      strcpy (output, "eight");
      break;
    case 9:
      strcpy (output, "nine");
    case 10:
      strcpy (output, "ten");
    default:
      sprintf (output, "%d", i);
    }
  return (output);
}


/*************************************************************************
 * openin_cdstream() -- Hugo Rabson                                      *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
int
openin_cdstream (struct s_bkpinfo *bkpinfo)
{
  return (openin_tape (bkpinfo));
}



/*************************************************************************
 * openin_tape() -- Hugo Rabson                                          *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
int
openin_tape (struct s_bkpinfo *bkpinfo)
{
	/** buffer ******************************************************/
  char fname[MAX_STR_LEN];
  char *datablock;
  char tmp[MAX_STR_LEN];
  char old_cwd[MAX_STR_LEN];
  char command[MAX_STR_LEN];
  char outfname[MAX_STR_LEN];
	/** int ********************************************************/
  int i;
  int j;
  int res;
  long length, k;
  int retval = 0;
  int ctrl_chr;

	/** long long **************************************************/
  long long size;

	/** pointers ***************************************************/
  FILE *fout;

	/** end vars ****************************************************/

  if (g_tape_stream) { log_it("FYI - I won't 'openin' the tape. It's already open."); return(0); }

  sprintf (outfname, "%s/tmp/all.tar.gz", bkpinfo->tmpdir);
  make_hole_for_file (outfname);
  if (bkpinfo->using_cdstream)
    {
      sprintf (command, "buffer -i %s -m 1m -p 50 -s %ld 2>> %s", bkpinfo->media_device, TAPE_BLOCK_SIZE, MONDO_LOGFILE);	/* -u 100 */
    }
  else
    {
      sprintf (command, "buffer -i %s %s -s %ld 2>> %s", bkpinfo->media_device, BUFFER_EXE_STUFF, TAPE_BLOCK_SIZE, MONDO_LOGFILE);	/* -u 100 */
    }
  log_it ("Opening IN stream with the command");
  log_it (command);
/*  log_it("Let's see what happens, shall we?"); */
  g_tape_stream = popen (command, "r");
  if (!g_tape_stream)
    {
      fatal_error ("Cannot openin stream device");
    }

  log_to_screen ("Reading stream");
  g_current_media_number = 1;
  sprintf (tmp, "stream device = '%s'", bkpinfo->media_device);
  log_it (tmp);
  g_tape_posK = 0;
/* skip data disks */
  open_evalcall_form ("Skipping data disks on stream");
  log_to_screen ("Skipping data disks on stream");
  if (!(fout = fopen (outfname, "w")))
    {
      fatal_error ("Cannot openout datadisk all.tar.gz file");
    }
  if (!(datablock = malloc(256*1024))) { fatal_error ("Unable to malloc 256*1024"); }
  for (i = 0; i < 32; i++)
    {
      for (j = 0; j < 4; j++)
	{
          for(length=k=0; length<256*1024; length+=k)
            {
	      k=fread(datablock+length, 1, 256 * 1024 - length, g_tape_stream);
            }
          fwrite (datablock, 1, length, fout);
	  g_tape_posK += length/1024;
	}
      update_evalcall_form ((i * 4 + j) * 100 / 128);
    }
  fclose (fout);
  free(datablock);
  datablock=NULL;
/* find initial blocks */
  res = read_header_block_from_tape (&size, fname, &ctrl_chr);
  retval += res;
  if (ctrl_chr != BLK_START_OF_TAPE)
    {
      wrong_marker (BLK_START_OF_BACKUP, ctrl_chr);
    }
  res = read_header_block_from_tape (&size, fname, &ctrl_chr);
  retval += res;
  if (ctrl_chr != BLK_START_OF_BACKUP)
    {
      wrong_marker (BLK_START_OF_BACKUP, ctrl_chr);
    }
  close_evalcall_form ();
  sprintf (tmp, "Saved all.tar.gz to '%s'", outfname);
  log_it (tmp);
/*
  sprintf(tmp,"cp -f %s /tmp", outfname);
  system(tmp);
*/
  getcwd (old_cwd, MAX_STR_LEN);
  chdir (bkpinfo->tmpdir);
  sprintf (tmp, "tar -zxf %s tmp/mondo-restore.cfg", outfname);
  run_program_and_log_output (tmp);
  system ("cp -f tmp/mondo-restore.cfg . 2> /dev/null");
  chdir (old_cwd);
  unlink (outfname);
  return (retval);
}




/*************************************************************************
 * openout_cdstream() -- Hugo Rabson                                     *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
int
openout_cdstream (char *cddev, int speed)
{
	/** buffers ******************************************************/
  char command[MAX_STR_LEN];

	/** end vars ****************************************************/

/*  add 'dummy' if testing */
  sprintf (command,
	   "cdrecord -eject dev=%s speed=%d fs=24m -waiti - >> %s 2>> %s",
	   cddev, speed, MONDO_LOGFILE, MONDO_LOGFILE);
  log_it ("Opening OUT cdstream with the command");
  log_it (command);
/*  log_it("Let's see what happens, shall we?"); */
  g_tape_stream = popen (command, "w");
  if (g_tape_stream)
    {
      return (0);
    }
  else
    {
      log_to_screen ("Failed to openout to cdstream (fifo)");
      return (1);
    }
}


/*************************************************************************
 * openout_tape() -- Hugo Rabson                                         *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
int
openout_tape (char *tapedev)
{
	/** buffer *******************************************************/
  char command[MAX_STR_LEN];


	/** end vars ****************************************************/

  if (g_tape_stream) { log_it("FYI - I won't 'openout' the tape. It's already open."); return(0); }

  sprintf (command, "buffer -o %s %s -s %d", tapedev, BUFFER_EXE_STUFF,
	   TAPE_BLOCK_SIZE);
  log_it ("Opening OUT tape with the command");
  log_it (command);
/*  log_it("Let's see what happens, shall we?"); */
  g_tape_stream = popen (command, "w");
  if (g_tape_stream)
    {
      return (0);
    }
  else
    {
      log_to_screen ("Failed to openout to tape (fifo)");
      return (1);
    }
}


/*************************************************************************
 * read_file_form_tape_to_file() -- Hugo Rabson                          *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
int
read_file_from_tape_to_file (struct s_bkpinfo *bkpinfo, char *outfile,
			     long long size)
{

	/** int *********************************************************/
  int res;

	/** end vars ****************************************************/

  res = read_file_from_tape_FULL (bkpinfo, outfile, NULL, size);

  return (res);
}




/*************************************************************************
 * read_file_from_tape_to_stream() -- Hugo Rabson                        *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
int
read_file_from_tape_to_stream (struct s_bkpinfo *bkpinfo, FILE * fout,
			       long long size)
{

	/** int *********************************************************/
  int res;

	/** end vars ****************************************************/

  res = read_file_from_tape_FULL (bkpinfo, NULL, fout, size);
/*  fflush(g_tape_stream);
  fflush(fout);*/
  return (res);
}

/*************************************************************************
 * read_file_from_tape_FULL() -- Hugo Rabson                             *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
int
read_file_from_tape_FULL (struct s_bkpinfo *bkpinfo, char *outfname,
			  FILE * foutstream, long long orig_size)
{
	/** buffers ******************************************************/
  char tmp[MAX_STR_LEN];
  char datablock[TAPE_BLOCK_SIZE];
  char temp_fname[MAX_STR_LEN];
  char temp_cksum[MAX_STR_LEN];
  char actual_cksum[MAX_STR_LEN];

	/** int **********************************************************/
  int retval = 0;
  int noof_blocks;
  int ctrl_chr;
  int res;

	/** pointers *****************************************************/
  FILE *fout;

	/** long    ******************************************************/
  long bytes_to_write = 0 /*,i */ ;

	/** long long ****************************************************/
  long long temp_size, size;

	/** unsigned int *************************************************/
  /*  unsigned int ch; */
  unsigned int crc16;
  unsigned int crctt;

	/** init  ********************************************************/
  crc16 = 0;
  crctt = 0;
  size = orig_size;


	/** end vars ****************************************************/
  res = read_header_block_from_tape (&temp_size, temp_fname, &ctrl_chr);
  if (orig_size != temp_size)
    {
      sprintf (tmp,
	       "output file's size should be %ld K but is apparently %ld K",
	       (long) size >> 10, (long) temp_size >> 10);
      log_to_screen (tmp);
    }
  if (ctrl_chr == BLK_END_OF_TAPE)
    {
      start_to_read_from_next_tape (bkpinfo);
      res = read_header_block_from_tape (&temp_size, temp_fname, &ctrl_chr);
    }
  if (ctrl_chr != BLK_START_FILE)
    {
      wrong_marker (BLK_START_FILE, ctrl_chr);
    }
  sprintf (tmp, "Reading file from tape; writing to '%s'; %ld KB", outfname,
	   (long) size >> 10);

  if (foutstream)
    {
      fout = foutstream;
    }
  else
    {
      fout = fopen (outfname, "w");
    }
  if (!fout)
    {
      log_it (outfname);
      fatal_error ("Cannot openout file");
    }
  for (noof_blocks = 0; size > 0; noof_blocks++, size -= bytes_to_write)
    {
      if (size < TAPE_BLOCK_SIZE)
	{
	  bytes_to_write = size;
	}
      else
	{
	  bytes_to_write = TAPE_BLOCK_SIZE;
	}
      g_tape_posK += fread (datablock, 1, /*bytes_to_write */ TAPE_BLOCK_SIZE,
			    g_tape_stream) / 1024;
      fwrite (datablock, 1, bytes_to_write, fout);
      /*
         for(i=0;i<bytes_to_write;i++)
         {
         ch=datablock[i];
         crc16=updcrcr(crc16,ch);
         crctt=updcrc(crctt,ch);
         }
       */
    }
  sprintf (actual_cksum, "%04x%04x", crc16, crctt);
  if (foutstream)
    {				/*log_it("Finished writing to foutstream"); */
    }
  else
    {
      fclose (fout);
    }
  res = read_header_block_from_tape (&temp_size, temp_cksum, &ctrl_chr);
  if (ctrl_chr != BLK_STOP_FILE)
    {
      wrong_marker (BLK_STOP_FILE, ctrl_chr);
    }
  if (strcmp (temp_cksum, actual_cksum))
    {
      sprintf (tmp, "actual cksum=%s; recorded cksum=%s", actual_cksum,
	       temp_cksum);
      log_to_screen (tmp);
      sprintf (tmp, "%s (%ld K) is corrupt on tape", temp_fname,
	       (long) orig_size >> 10);
      fatal_error (tmp);
    }
  else
    {
      sprintf (tmp, "%s is GOOD on tape", temp_fname);
      /*      log_it(tmp); */
    }
  return (retval);
}



/*************************************************************************
 * read_header_block_from_tape() -- Hugo Rabson                          *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
int
read_header_block_from_tape (long long *plen, char *filename,
			     int *pcontrol_char)
{

	/** buffers ******************************************************/
  char tempblock[TAPE_BLOCK_SIZE];
  char tmp[MAX_STR_LEN];

	/** int **********************************************************/
  int i;

	/** end vars ****************************************************/

  for (i = 0; i < TAPE_BLOCK_SIZE; i++)
    {
      tempblock[i] = 0;
    }
  g_tape_posK += fread (tempblock, 1, TAPE_BLOCK_SIZE, g_tape_stream) / 1024;
  *pcontrol_char = tempblock[7000];
/*  memcpy((char*)plength_of_incoming_file,(char*)tempblock+7001,sizeof(long long)); */
/*  for(*plen=0,i=7;i>=0;i--) {*plen<<=8; *plen |= tempblock[7001+i];} */
  memcpy ((char *) plen, tempblock + 7001, sizeof (long long));
  if (strcmp (tempblock + 6000 + *pcontrol_char, "Mondolicious, baby"))
    {
      sprintf (tmp, "Bad header block at %ld K", (long) g_tape_posK);
      log_it (tmp);
    }
  strcpy (filename, tempblock + 1000);
/*  strcpy(cksum,tempblock+5555);*/
  sprintf (tmp, "%s  (reading) fname=%s, filesize=%ld K",
	   marker_to_string (*pcontrol_char), filename,
	   (long) ((*plen) >> 10));
/*  log_it(tmp); */
  if (*pcontrol_char == BLK_ABORTED_BACKUP)
    {
      log_to_screen ("I can't verify an aborted backup.");
      finish (1);
    }
  return (0);
}




/*************************************************************************
 * reset_bkpinfo() -- Hugo Rabson                                        *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
void
reset_bkpinfo (struct s_bkpinfo *bkpinfo)
{
  bkpinfo->manual_cd_tray = FALSE;
  bkpinfo->media_device[0] = '\0';
  bkpinfo->media_size[0] = -1;
  bkpinfo->boot_loader = '\0';
  bkpinfo->boot_device[0] = '\0';
  bkpinfo->zip_exe[0] = '\0';
  bkpinfo->zip_suffix[0] = '\0';
  bkpinfo->restore_path[0] = '\0';
  bkpinfo->compression_level = 2;	/* default */
  bkpinfo->use_lzo = FALSE;
  bkpinfo->do_not_compress_these[0] = '\0';
  bkpinfo->verify_data = FALSE;
  bkpinfo->backup_data = FALSE;
  bkpinfo->isodir[0] = '\0';
  bkpinfo->scratchdir[0] = '\0';
  bkpinfo->tmpdir[0] = '\0';
  bkpinfo->optimal_set_size = 0;
  bkpinfo->using_tape = FALSE;
  bkpinfo->include_paths[0] = '\0';
  bkpinfo->exclude_paths[0] = '\0';
  bkpinfo->call_before_iso[0] = '\0';
  bkpinfo->call_make_iso[0] = '\0';
  bkpinfo->call_burn_iso[0] = '\0';
  bkpinfo->call_after_iso[0] = '\0';
  bkpinfo->image_devs[0] = '\0';
  bkpinfo->postnuke_tarball[0] = '\0';
  bkpinfo->kernel_path[0] = '\0';
  bkpinfo->nfs_mount[0] = '\0';
  bkpinfo->nfs_remote_dir[0] = '\0';
  bkpinfo->wipe_media_first = FALSE;
  bkpinfo->differential = FALSE;
  bkpinfo->cdrw_speed = 0;
}




/*************************************************************************
 * reload_filelist() -- Hugo Rabson                                      *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
void
reload_filelist (struct s_node *filelist)
{
  toggle_node_selection (filelist, FALSE);
  toggle_path_expandability (filelist, "/", FALSE);
  toggle_all_root_dirs_on (filelist);
}




/*************************************************************************
 * run_program_and_log_output() -- Hugo Rabson                           *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
int
run_program_and_log_output (char *program)
{
	/** buffer *******************************************************/
  char callstr[MAX_STR_LEN];
  char incoming[MAX_STR_LEN];
  char tmp[MAX_STR_LEN];

	/** int **********************************************************/
  int res;
  int i;
  int len;

	/** pointers ****************************************************/
  FILE *fin;
  char *p;

	/** end vars ****************************************************/
  sprintf (callstr, "%s > /tmp/mondo-run-prog.tmp 2> /tmp/mondo-run-prog.err",
	   program);
  while ((p = strchr (callstr, '\r')))
    {
      *p = ' ';
    }
  while ((p = strchr (callstr, '\n')))
    {
      *p = ' ';
    }				/* single '=' is intentional */
  res = system (callstr);
  len = strlen (program);
  for (i = 0; i < 35 - len / 2; i++)
    {
      tmp[i] = '-';
    }
  tmp[i] = '\0';
  strcat (tmp, " ");
  strcat (tmp, program);
  strcat (tmp, " ");
  for (i = 0; i < 35 - len / 2; i++)
    {
      strcat (tmp, "-");
    }
  log_it (tmp);
  system ("cat /tmp/mondo-run-prog.err >> /tmp/mondo-run-prog.tmp");
  unlink ("/tmp/mondo-run-prog.err");
  fin = fopen ("/tmp/mondo-run-prog.tmp", "r");
  if (fin)
    {
      for (fgets (incoming, MAX_STR_LEN, fin); !feof (fin);
	   fgets (incoming, MAX_STR_LEN, fin))
	{
/*	  for(i=0; incoming[i]!='\0'; i++);
	  for(; i>0 && incoming[i-1]<=32; i--);
	  incoming[i]='\0';
*/
	  strip_spaces (incoming);
	  log_it (incoming);
	}
      fclose (fin);
    }
  log_it
    ("--------------------------------end of output------------------------------");
  if (res)
    {
      log_it ("...ran with errors.");
    }
  /* else { log_it("...ran just fine. :-)"); } */
  return (res);
}




/*************************************************************************
 * run_program_and_log_to_screen() -- Hugo Rabson                        *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
int
run_program_and_log_to_screen (char *basic_call, char *what_i_am_doing)
{
	/** int *********************************************************/
  int retval = 0;
  int res = 0;
  int i;

	/**pointers *****************************************************/
  FILE *fin;

	/** buffers *****************************************************/
  char tmp[MAX_STR_LEN];
  char command[MAX_STR_LEN];
  char lockfile[MAX_STR_LEN];

	/** end vars ****************************************************/
  sprintf (lockfile,
	   call_program_and_get_last_line_of_output
	   ("mktemp -q /tmp/blah.XXXXXX"));
  sprintf (command,
	   "echo hi > %s ; %s >> %s 2>> %s; res=$?; rm -f %s; exit $res",
	   lockfile, basic_call, MONDO_LOGFILE, MONDO_LOGFILE, lockfile);
  open_evalcall_form (what_i_am_doing);
  sprintf (tmp, "Executing %s", basic_call);
  log_it (tmp);
  fin = popen (command, "r");
  if (!fin)
    {
      sprintf (tmp, "Failed utterly to call '%s'", command);
      log_to_screen (tmp);
      return (1);
    }
  for (i = 0; i < 3 && !does_file_exist (lockfile); sleep (1), i++)
    {
/*      log_it("So tired.. Tired of waiting..."); */
    }
  for (; does_file_exist (lockfile); sleep (1))
    {
      log_file_end_to_screen (MONDO_LOGFILE, "");
      update_evalcall_form (0);
    }
  if (pclose (fin))
    {
      res = 1;
    }
  retval += res;
  close_evalcall_form ();
  return (retval);
}

/*************************************************************************
 * save_filelist() -- Hugo Rabson                                        *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
void
save_filelist (struct s_node *filelist, char *outfname)
{
	/** int **********************************************************/
  static int len = 0;
  static int percentage;
  static int depth = 0;

	/** buffers ******************************************************/
  static char str[MAX_STR_LEN];

	/** structures ***************************************************/
  struct s_node *node;

	/** pointers *****************************************************/
  static FILE *fout = NULL;

	/** long *********************************************************/
  static long lines_in_filelist = 0;
  static long lino = 0;

	/** end vars ****************************************************/

  if (depth == 0)
    {
      log_to_screen ("Saving filelist");
      fout = fopen (outfname, "w");
      lines_in_filelist = g_original_noof_lines_in_filelist;	/* set by load_filelist() */
      open_evalcall_form ("Saving selection to disk");
    }
  depth++;
  for (node = filelist; node != NULL; node = node->right)
    {
      str[len] = node->ch;
      if (!node->ch)
	{
	  if (node->selected)
	    {
	      fprintf (fout, "%s\n", str);
	    }
	  if (!(++lino % 1111))
	    {
	      percentage = (int) (lino * 100 / lines_in_filelist);
	      update_evalcall_form (percentage);
	    }
	}
      if (node->down)
	{
	  len++;
	  save_filelist (node->down, "");
	  len--;
	}
    }
  depth--;
  if (depth == 0)
    {
      fclose (fout);
      fout = NULL;
      close_evalcall_form ();
      log_it ("Finished saving filelist");
    }
}

/*************************************************************************
 * size_of_all_biggiefiles_K() -- Hugo Rabson                            *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
long
size_of_all_biggiefiles_K (struct s_bkpinfo *bkpinfo)
{
	/** buffers ******************************************************/
  char fname[MAX_STR_LEN];
  char biggielist[MAX_STR_LEN];
  char command[MAX_STR_LEN];
  char res_sz[MAX_STR_LEN];
  char comment[MAX_STR_LEN];

	/** long *********************************************************/
  long scratchL = 0;
  long file_len_K;

	/** pointers ****************************************************/
  FILE *fin;

	/** end vars ****************************************************/


  log_it ("Calculating size of all biggiefiles (in total)");
  sprintf (biggielist, "%s/biggielist.txt", bkpinfo->tmpdir);
  if (!(fin = fopen (biggielist, "r")))
    {
      log_it
	("Cannot open biggielist. OK, so estimate is based on filesets only.");
    }
  else
    {
      for (fgets (fname, MAX_STR_LEN, fin); !feof (fin);
	   fgets (fname, MAX_STR_LEN, fin))
	{
	  strip_spaces (fname);
	  if (0 == strncmp (fname, "/dev/", 5))
	    {
	      sprintf (command,
		       "cat %s/mountlist.txt | grep \"%s \" | head -n1 | cut -d' ' -f4",
		       bkpinfo->tmpdir, fname);
	      log_it (command);
	      strcpy (res_sz,
		      call_program_and_get_last_line_of_output (command));
	      file_len_K = atol (res_sz);
	      if (file_len_K > 0)
		{
		  scratchL += file_len_K;
		}
	    }
	  else
	    {
	      file_len_K = (long) (length_of_file (fname) / 1024);
	      if (file_len_K > 0)
		{
		  scratchL += file_len_K;
		}
	    }
	  sprintf (comment, "After adding %s, scratchL+%ld now equals %ld",
		   fname, file_len_K, scratchL);
/*          log_it(comment); */
	}
    }
  fclose (fin);
  log_it ("Finished calculating total size of all biggiefiles");
  return (scratchL);
}


/*************************************************************************
 * slice_fname() -- Hugo Rabson                                          *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
char *
slice_fname (long bigfileno, long sliceno, char *path, char *s)
{

	/** buffers *****************************************************/
  static char output[MAX_STR_LEN];
  static char suffix[MAX_STR_LEN];

	/** end vars ****************************************************/


  if (s[0] != '\0')
    {
      sprintf (suffix, ".%s", s);
    }
  else
    {
      suffix[0] = '\0';
    }
  sprintf (output, "%s/slice-%07ld.%05ld.dat%s", path, bigfileno, sliceno,
	   suffix);
  return (output);
}



/*************************************************************************
 * space_occupied_by_cd() -- Hugo Rabson                                 *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
long long
space_occupied_by_cd (char *mountpt)
{
	/** buffer *******************************************************/
  char tmp[MAX_STR_LEN];
  char command[MAX_STR_LEN];

	/** pointers *****************************************************/
  char *p;
  FILE *fin;

	/** end vars ****************************************************/

  sprintf (command, "du -sk %s", mountpt);
  fin = popen (command, "r");
  fgets (tmp, MAX_STR_LEN, fin);
  pclose (fin);
  p = strchr (tmp, '\t');
  if (p)
    {
      *p = '\0';
    }
  return (atoll (tmp));
}




/*************************************************************************
 * should_we_write_to_next_tape() -- Hugo Rabson                         *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
bool
should_we_write_to_next_tape (long mediasize,
			      long long length_of_incoming_file)
{
	/** bool's ******************************************************/
  bool we_need_a_new_tape = FALSE;

	/** end vars ****************************************************/
  if (g_tape_posK >> 10 >= mediasize)
    {
      we_need_a_new_tape = TRUE;
      log_to_screen ("Should have started a new tape/CD already");
    }
  if ((g_tape_posK + length_of_incoming_file / 1024) >> 10 >= mediasize)
    {
      we_need_a_new_tape = TRUE;
    }
  return (we_need_a_new_tape);
}


/*************************************************************************
 * start_to_read_from_next_tape() -- Hugo Rabson                         *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
int
start_to_read_from_next_tape (struct s_bkpinfo *bkpinfo)
{
	/** int **********************************************************/
  int res = 0;
  int ctrl_chr = '\0';

  /** buffers ************************************************************/
  char fname[MAX_STR_LEN];	/*cksum[MAX_STR_LEN]; */

  /** long long *********************************************************/
  long long size;

	/** end vars ****************************************************/

  pclose (g_tape_stream);
  g_tape_stream = NULL;
  log_it ("Next tape requested.");
  popup_and_OK ("Please insert the next tape/CD in this series.");
  g_current_media_number++;
  g_tape_stream = fopen (bkpinfo->media_device, "r");
  g_tape_posK = 0;
  res = read_header_block_from_tape (&size, fname, &ctrl_chr);
  if (ctrl_chr != BLK_START_OF_TAPE)
    {
      wrong_marker (BLK_START_OF_BACKUP, ctrl_chr);
    }
  return (res);
}




/*************************************************************************
 * start_to_write_to_next_tape() -- Hugo Rabson                          *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
int
start_to_write_to_next_tape (struct s_bkpinfo *bkpinfo)
{
  write_header_block_to_tape (g_current_media_number, "end-of-tape",
			      BLK_END_OF_TAPE);
  pclose (g_tape_stream);
  g_tape_stream = NULL;

  log_it ("New tape requested.");

  popup_and_OK ("This tape/CD is full. Please insert another.");

  g_current_media_number++;
  if (g_current_media_number > MAX_NOOF_MEDIA) { fatal_error("Too many tapes. Man, you need to use nfs!"); }

  openout_tape (bkpinfo->media_device);	/* sets g_tape_stream */

/*g_tape_stream = fopen(bkpinfo->media_device,"w"); */

  g_tape_posK = 0;

  write_header_block_to_tape (g_current_media_number, "start-of-tape",
			      BLK_START_OF_TAPE);
  return (0);
}


/*************************************************************************
 * strcmp_inc_numbers() -- Hugo Rabson                                   *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
int
strcmp_inc_numbers (char *stringA, char *stringB)
{
	/** int **********************************************************/
  int i;
  int start_of_numbers_in_A;
  int start_of_numbers_in_B;
  int res;

	/** long ********************************************************/
  long numA;
  long numB;

	/** end vars ****************************************************/

  if (strlen (stringA) == strlen (stringB))
    {
      return (strcmp (stringA, stringB));
    }
  for (i = strlen (stringA); i > 0 && isdigit (stringA[i - 1]); i--);
  if (i == strlen (stringA))
    {
      return (strcmp (stringA, stringB));
    }
  start_of_numbers_in_A = i;
  for (i = strlen (stringB); i > 0 && isdigit (stringB[i - 1]); i--);
  if (i == strlen (stringB))
    {
      return (strcmp (stringA, stringB));
    }
  start_of_numbers_in_B = i;
  if (start_of_numbers_in_A != start_of_numbers_in_B)
    {
      return (strcmp (stringA, stringB));
    }
  res = strncmp (stringA, stringB, i);
  if (res)
    {
      return (res);
    }
  numA = atol (stringA + start_of_numbers_in_A);
  numB = atol (stringB + start_of_numbers_in_B);
  /*
     sprintf(tmp,"Comparing %s and %s --> %ld,%ld\n",stringA,stringB,numA,numB);
     log_to_screen(tmp);
   */
  return (numA - numB);
}




/*************************************************************************
 *strip_afio_output_line() -- Hugo Rabson                                *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
char *
strip_afio_output_line (char *input)
{
	/** buffer *******************************************************/
  static char output[MAX_STR_LEN];

	/** pointers *****************************************************/
  char *p;
  char *q;
	/** end vars ****************************************************/


  strcpy (output, input);
  p = strchr (input, '\"');
  if (p)
    {
      q = strchr (++p, '\"');
      if (q)
	{
	  strcpy (output, p);
	  *(strchr (output, '\"')) = '\0';
	}
    }
  return (output);
}



/*************************************************************************
 * strip_spaces() -- Hugo Rabson                                         *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
void
strip_spaces (char *in_out)
{
	/** buffers ******************************************************/
  char tmp[MAX_STR_LEN];

	/** pointers *****************************************************/
  char *p;

	/** int *********************************************************/
  int i;

	/** end vars ****************************************************/
  for (i = 0; in_out[i] <= ' ' && i < strlen (in_out); i++);
  strcpy (tmp, in_out + i);
  for (i = strlen (tmp); i>0 && tmp[i - 1] <= 32; i--);
  tmp[i] = '\0';
  for (i = 0; i < 80; i++)
    {
      in_out[i] = ' ';
    }
  in_out[i] = '\0';
  i = 0;
  p = tmp;
  while (*p != '\0')
    {
      in_out[i] = *(p++);
      in_out[i + 1] = '\0';
      if (in_out[i] < 32 && i > 0)
	{
	  if (in_out[i] == 8)
	    {
	      i--;
	    }
	  else if (in_out[i] == 9)
	    {
	      in_out[i++] = ' ';
	    }
	  else if (in_out[i] == '\t')
	    {
	      for (i++; i % 5; i++);
	    }
	  else if (in_out[i] >= 10 && in_out[i] <= 13)
	    {
	      break;
	    }
	  else
	    {
	      i--;
	    }
	}
      else
	{
	  i++;
	}
    }
  in_out[i] = '\0';
/*  for(i=strlen(in_out); i>0 && in_out[i-1]<=32; i--) {in_out[i-1]='\0';} */
}




/*************************************************************************
 * toggle_all_root_dirs_on() -- Hugo Rabson                              *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
void
toggle_all_root_dirs_on (struct s_node *filelist)
{
	/** structures ***************************************************/
  struct s_node *node;

	/** int **********************************************************/
  static int depth = 0;
  static int root_dirs_expanded;

	/** buffers ******************************************************/
  static char filename[MAX_STR_LEN];

	/** end vars ****************************************************/
  if (depth == 0)
    {
      log_it ("Toggling all root dirs ON");
      root_dirs_expanded = 0;
    }
  for (node = filelist; node != NULL; node = node->right)
    {
      filename[depth] = node->ch;
      if (node->ch == '\0' && strlen (filename) > 1
	  && (!strchr (filename + 1, '/')))
	{
	  node->selected = FALSE;
	  node->expanded = TRUE;
	  log_it (filename);
	  root_dirs_expanded++;
	}
      if (node->down)
	{
	  depth++;
	  toggle_all_root_dirs_on (node->down);
	  depth--;
	}
    }
  if (depth == 0)
    {
      log_it ("Finished toggling all root dirs ON");
    }
}



/*************************************************************************
 * toggle_path_expandability() -- Hugo Rabson                            *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
void
toggle_path_expandability (struct s_node *filelist, char *pathname,
			   bool on_or_off)
{

	/** int *********************************************************/
  static int depth = 0;
  static int total_expanded;
  static int root_depth;
  int j;
	/** structures **************************************************/
  struct s_node *node;

	/** buffers *****************************************************/
  static char current_filename[MAX_STR_LEN];

/*  char tmp[MAX_STR_LEN+2]; */

	/** end vars ****************************************************/
  if (depth == 0)
    {
      total_expanded = 0;
      log_it ("Toggling path's expandability");
      for (root_depth = strlen (pathname);
	   root_depth > 0 && pathname[root_depth - 1] != '/'; root_depth--);
      if (root_depth < 2)
	{
	  root_depth = strlen (pathname);
	}
    }
  for (node = filelist; node != NULL; node = node->right)
    {
      current_filename[depth] = node->ch;
      if (node->down)
	{
	  depth++;
	  toggle_path_expandability (node->down, pathname, on_or_off);
	  depth--;
	}
      if (node->ch == '\0')
	{
	  if (!strncmp (pathname, current_filename, strlen (pathname)))
	    {
	      for (j = root_depth;
		   current_filename[j] != '/' && current_filename[j] != '\0';
		   j++);
	      if (current_filename[j] != '\0')
		{
		  for (j++;
		       current_filename[j] != '/'
		       && current_filename[j] != '\0'; j++);
		}
	      if (current_filename[j] == '\0')
		{
/*                  sprintf(tmp,"j=%d; root_depth=%d; stem(path)=%s; stem(curr)=%s",j,root_depth,pathname+j, current_filename+j);
                  log_it(tmp); */
		  node->expanded =
		    (!strcmp (pathname, current_filename) ? TRUE : on_or_off);
		}
	    }
	}
      if (node->expanded)
	{
	  if (total_expanded < ARBITRARY_MAXIMUM - 32
	      || !strrchr (current_filename + strlen (pathname), '/'))
	    {
	      total_expanded++;
	    }
	  else
	    {
	      node->expanded = FALSE;
	    }
	}
    }
  if (depth == 0)
    {
      log_it ("Finished toggling expandability");
    }
}


/*************************************************************************
 * toggle_path_selection() -- Hugo Rabson                                *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
void
toggle_path_selection (struct s_node *filelist, char *pathname,
		       bool on_or_off)
{
	/** int **********************************************************/
  static int depth = 0;
  int j;

	/** structures ***************************************************/
  struct s_node *node;

	/** buffers ******************************************************/
  static char current_filename[MAX_STR_LEN];
  char tmp[MAX_STR_LEN + 2];

	/** end vars ****************************************************/
  if (depth == 0)
    {
      log_it ("Toggling path's selection");
    }
  for (node = filelist; node != NULL; node = node->right)
    {
      current_filename[depth] = node->ch;
      if (node->down)
	{
	  depth++;
	  toggle_path_selection (node->down, pathname, on_or_off);
	  depth--;
	}
      if (node->ch == '\0')
	{
	  if (!strncmp (pathname, current_filename, strlen (pathname)))
	    {
	      for (j = 0;
		   pathname[j] != '\0' && pathname[j] == current_filename[j];
		   j++);
	      if (current_filename[j] == '/' || current_filename[j] == '\0')
		{
		  node->selected = on_or_off;
		  sprintf (tmp, "%s is now %s\n", current_filename,
			   (on_or_off ? "ON" : "OFF"));
		}
	    }
	}
    }
  if (depth == 0)
    {
      log_it ("Finished toggling selection");
    }
}



/*************************************************************************
 * toggle_node_selection() -- Hugo Rabson                                *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
void
toggle_node_selection (struct s_node *filelist, bool on_or_off)
{
	/** structure ***************************************************/
  struct s_node *node;

	/** end vars ****************************************************/
  for (node = filelist; node != NULL; node = node->right)
    {
      if (node->ch == '/')
	{
	  continue;
	}			/* don't go deep */
      if (node->ch == '\0')
	{
	  node->selected = on_or_off;
	}
      if (node->down)
	{
	  toggle_node_selection (node->down, on_or_off);
	}
    }
}









/*-----------------------------------------------------------*/
/*************************************************************************
 * trim_empty_quotes() -- Hugo Rabson                                    *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
char *
trim_empty_quotes (char *incoming)
{
	/** buffer *******************************************************/
  static char outgoing[MAX_STR_LEN];

	/** end vars ****************************************************/
  if (incoming[0] == '\"' && incoming[strlen (incoming) - 1] == '\"')
    {
      strcpy (outgoing, incoming + 1);
      outgoing[strlen (outgoing) - 1] = '\0';
    }
  else
    {
      strcpy (outgoing, incoming);
    }
  return (outgoing);
}


/*************************************************************************
 * which_boot_loader() -- Hugo Rabson                                    *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
char
which_boot_loader (char *which_device)
{
	/** buffer ******************************************************/
  char list_drives_cmd[MAX_STR_LEN];
  char tmp[MAX_STR_LEN];
  char current_drive[MAX_STR_LEN];

	/** pointers ****************************************************/
  FILE *pdrives;

	/** int *********************************************************/
  int count_lilos = 0;
  int count_grubs = 0;

	/** end vars ****************************************************/
  sprintf (list_drives_cmd,
	   "fdisk -l | grep /dev | grep cyl | tr ':' ' ' | cut -d' ' -f2");
  if (!(pdrives = popen (list_drives_cmd, "r")))
    {
      fatal_error ("Unable to open list of drives");
    }
  for (fgets (current_drive, MAX_STR_LEN, pdrives); !feof (pdrives);
       fgets (current_drive, MAX_STR_LEN, pdrives))
    {
      strip_spaces (current_drive);
      sprintf (tmp, "looking at drive %s's MBR", current_drive);
      log_it (tmp);
      if (does_string_exist_in_boot_block (current_drive, "GRUB"))
	{
	  count_grubs++;
	  strcpy (which_device, current_drive);
	  break;
	}
      if (does_string_exist_in_boot_block (current_drive, "LILO"))
	{
	  count_lilos++;
	  strcpy (which_device, current_drive);
	  break;
	}
    }
  pclose (pdrives);
  sprintf (tmp, "%d grubs and %d lilos\n", count_grubs, count_lilos);
  log_it (tmp);
  if (count_grubs && !count_lilos)
    {
      return ('G');
    }
  else if (count_lilos && !count_grubs)
    {
      return ('L');
    }
  else if (count_grubs == 1 && count_lilos == 1)
    {
      log_it ("I'll bet you used to use LILO but switched to GRUB...");
      return ('G');
    }
  else
    {
      log_it ("Unknown boot loader");
      return ('U');
    }
}


/*************************************************************************
 * read_cfg_var() -- Hugo Rabson                                         *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
int
read_cfg_var (char *config_file, char *label, char *value)
{
	/** buffer *******************************************************/
  char command[MAX_STR_LEN];
  char tmp[MAX_STR_LEN];

	/** end vars ****************************************************/
  if (!does_file_exist (config_file))
    {
      sprintf (tmp, "(read_cfg_var) Cannot find %s config file", config_file);
      log_to_screen (tmp);
      value[0] = '\0';
      return (-1);
    }
  else
    {
      sprintf (command, "cat %s | grep \"%s .*\" | cut -d' ' -f2,3,4,5",
	       config_file, label);
      strcpy (value, call_program_and_get_last_line_of_output (command));
      if (strlen (value) == 0)
	{
	  return (1);
	}
      else
	{
	  return (0);
	}
    }
}


/*************************************************************************
 * write_cfg_var() -- Hugo Rabson                                        *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
int
write_cfg_var (char *config_file, char *label, char *value)
{
	/** buffers ******************************************************/
  char command[MAX_STR_LEN];
  char tempfile[MAX_STR_LEN];
  char tmp[MAX_STR_LEN];


	/** end vars ****************************************************/
  if (!does_file_exist (config_file))
    {
      sprintf (tmp, "(write_cfg_file) Cannot find %s config file",
	       config_file);
      fatal_error (tmp);
    }
  strcpy (tempfile,
	  call_program_and_get_last_line_of_output
	  ("mktemp -q /tmp/blah.XXXXXX"));
  if (does_file_exist (config_file))
    {
      sprintf (command, "cat %s | grep -vx \"%s .*\" > %s", MONDO_CFG_FILE,
	       label, tempfile);
      system (command);
    }
  sprintf (command, "echo \"%s %s\" >> %s", label, value, tempfile);
  system (command);
  sprintf (command, "mv -f %s %s", tempfile, MONDO_CFG_FILE);
  system (command);
  return (0);
}




/*************************************************************************
 * write_data_disk_to_tape() -- Hugo Rabson                              *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
int
write_data_disks_to_tape (char *fname)
{
	/** pointers ****************************************************/
  FILE *fin;

	/** long ********************************************************/
  long m = -1;

	/** int *********************************************************/
  int i;
  int j;

	/** buffers *****************************************************/
  char tempblock[256 * 1024];
  char tmp[MAX_STR_LEN];

	/** end vars ****************************************************/

  open_evalcall_form ("Writing data disks to tape");
  log_to_screen ("Writing data disks to tape");
  if (!does_file_exist (fname))
    {
      sprintf (tmp, "Cannot find %s", fname);
      log_to_screen (tmp);
      finish (1);
    }
  if (!(fin = fopen (fname, "r")))
    {
      fatal_error ("Cannot openin the data disk");
    }
  for (i = 0; i < 32; i++)	/* 32MB */
    {
      for (j = 0; j < 4; j++)	/* 256K x 4 = 1MB (1024K) */
	{
	  if (!feof (fin))
	    {
	      m = fread (tempblock, 1, 256 * 1024, fin);
	    }
	  else
	    {
	      m = 0;
	    }
	  for (; m < 256 * 1024; m++)
	    {
	      tempblock[m] = '\0';
	    }
	  g_tape_posK +=
	    fwrite (tempblock, 1, 256 * 1024, g_tape_stream) / 1024;
	}
      update_evalcall_form ((i * 4 + j) * 100 / 128);
    }
  fclose (fin);
  close_evalcall_form ();
  return (0);
}



/*************************************************************************
 * write_file_to_tape_from_file() -- Hugo Rabson                         *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
int
write_file_to_tape_from_file (struct s_bkpinfo *bkpinfo, char *infile)
{
	/** buffers *****************************************************/
  char tmp[MAX_STR_LEN];
  char datablock[TAPE_BLOCK_SIZE];
  char checksum[MAX_STR_LEN];

	/** int *********************************************************/
  int retval = 0;
  int noof_blocks;

  /*  unsigned int ch; */
  unsigned int crc16;
  unsigned int crctt;

	/** pointers ****************************************************/
  FILE *fin;
  char *p;

	/** long ********************************************************/
  long bytes_to_read = 0;
  long i;

	/** long long ***************************************************/
  long long filesize;

	/** initialize *************************************************/
  crc16 = 0;
  crctt = 0;

	/** end vars ****************************************************/
  filesize = length_of_file (infile);
  if (should_we_write_to_next_tape (bkpinfo->media_size[g_current_media_number], filesize))
    {
      start_to_write_to_next_tape (bkpinfo);
    }
  p = strrchr (infile, '/');
  if (!p)
    {
      p = infile;
    }
  else
    {
      p++;
    }
  sprintf (tmp, "Writing file '%s' to tape (%ld KB)", p,
	   (long) filesize >> 10);
  log_it (tmp);
  write_header_block_to_tape (filesize, infile, BLK_START_FILE);
  fin = fopen (infile, "r");
  for (noof_blocks = 0; filesize > 0;
       noof_blocks++, filesize -= bytes_to_read)
    {
      if (filesize < TAPE_BLOCK_SIZE)
	{
	  bytes_to_read = filesize;
	  for (i = 0; i < TAPE_BLOCK_SIZE; i++)
	    {
	      datablock[i] = '\0';
	    }
	}
      else
	{
	  bytes_to_read = TAPE_BLOCK_SIZE;
	}
      fread (datablock, 1, bytes_to_read, fin);
      g_tape_posK += fwrite (datablock, 1, /*bytes_to_read */ TAPE_BLOCK_SIZE,
			     g_tape_stream) / 1024;
      /*
         for(i=0;i<bytes_to_read;i++)
         {
         ch=datablock[i];
         crc16=updcrcr(crc16,ch);
         crctt=updcrc(crctt,ch);
         }
       */
    }
  fclose (fin);
  sprintf (checksum, "%04x%04x", crc16, crctt);
  write_header_block_to_tape (g_current_media_number, checksum,
			      BLK_STOP_FILE);
  sprintf (tmp, "File '%s' written to tape.", infile);
  log_it (tmp);
  return (retval);
}




/*************************************************************************
 * write_one_liner_data_file() -- Hugo Rabson                            *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
int
write_one_liner_data_file (char *fname, char *contents)
{
	/** pointers ****************************************************/
  FILE *fout;

	/** end vars ****************************************************/

  if (!(fout = fopen (fname, "w")))
    {
      return (1);
    }
  fprintf (fout, "%s\n", contents);
  fclose (fout);
  return (0);
}






/*************************************************************************
 *write_header_block_to_tape() -- Hugo Rabson                            *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
int
write_header_block_to_tape (long long length_of_incoming_file, char *filename,
			    int control_char)
{
	/** buffers *****************************************************/
  char tempblock[TAPE_BLOCK_SIZE];
  char tmp[MAX_STR_LEN];
  char *p;

	/** int *********************************************************/
  int i;

	/** long long ***************************************************/
  long long olen;

	/** end vars ****************************************************/


  olen = length_of_incoming_file;
  p = strrchr (filename, '/');	/* Make 'em go, "Unnnh!" Oh wait, that was _Master_ P... */
  if (!p)
    {
      p = filename;
    }
  else
    {
      p++;
    }
  if (!g_tape_stream)
    {
      log_to_screen
	("You're not backing up to tape. Why write a tape header?");
      finish (1);
    }
  for (i = 0; i < TAPE_BLOCK_SIZE; i++)
    {
      tempblock[i] = 0;
    }
  sprintf (tempblock + 6000 + control_char, "Mondolicious, baby");
  tempblock[7000] = control_char;
/*  for(i=0;i<8;i++) {tempblock[7001+i]=olen&0xff; olen>>=8;} */
  memcpy (tempblock + 7001, (char *) &olen, sizeof (long long));
/*  if (length_of_incoming_file) {memcpy(tempblock+7001,(char*)&length_of_incoming_file,sizeof(long long));} */
  strcpy (tempblock + 1000, filename);
/*  strcpy(tempblock+5555,cksum); */
  g_tape_posK += fwrite (tempblock, 1, TAPE_BLOCK_SIZE, g_tape_stream) / 1024;
  sprintf (tmp, "%s (fname=%s, size=%ld K)", marker_to_string (control_char),
	   p, (long) length_of_incoming_file >> 10);
  /*  log_it(tmp); */
/*  log_tape_pos(); */
  return (0);
}



/*************************************************************************
 * wrong_marker() -- Hugo Rabson                                         *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
void
wrong_marker (int should_be, int it_is)
{
	/** buffer ******************************************************/
  char tmp[MAX_STR_LEN];


	/** end vars ****************************************************/
  sprintf (tmp, "Wrong marker! (Should be %s, ",
	   marker_to_string (should_be));
  sprintf (tmp + strlen (tmp), "is actually %s)", marker_to_string (it_is));
  fatal_error (tmp);
}




/*************************************************************************
 * zero_out_a_device() -- Hugo Rabson                                    *
 *                                                                       *
 * Purpose:                                                              *
 * Called by:                                                            *
 * Returns:                                                              *
 *************************************************************************/
int
zero_out_a_device (char *device)
{
	/** pointer *****************************************************/
  FILE *fout;

	/** buffers *****************************************************/
  char tmp[MAX_STR_LEN];

	/** int *********************************************************/
  int i;

	/** end vars ****************************************************/
  sprintf (tmp, "Zeroing drive %s", device);
  log_it (tmp);
  fout = fopen (device, "w");
  if (!fout)
    {
      log_it ("Unable to open/write to device");
      return (1);
    }
  for (i = 0; i < 16384; i++)
    {
      fputc ('\0', fout);
    }
  fclose (fout);
  log_it ("Device successfully zeroed.");
  return (0);
}



/* end of mondo-tools.c */














